"""This module defines SequencePulseTemplate, a higher-order hierarchical pulse template that
combines several other PulseTemplate objects for sequential execution."""

import numpy as np
from typing import Dict, List, Tuple, Set, Optional, Any, Iterable, Union
from numbers import Real
import functools
import warnings

from cached_property import cached_property

from qupulse.serialization import Serializer, PulseRegistryType
from qupulse._program._loop import Loop

from qupulse.utils.types import MeasurementWindow, ChannelID, TimeType
from qupulse.pulses.pulse_template import PulseTemplate
from qupulse.pulses.parameters import Parameter, ParameterConstrainer, ParameterNotProvidedException
from qupulse.pulses.sequencing import InstructionBlock, Sequencer
from qupulse.pulses.conditions import Condition
from qupulse.pulses.mapping_pulse_template import MappingPulseTemplate, MappingTuple
from qupulse._program.waveforms import SequenceWaveform
from qupulse.pulses.measurement import MeasurementDeclaration, MeasurementDefiner
from qupulse.expressions import Expression, ExpressionScalar

__all__ = ["SequencePulseTemplate"]


class SequencePulseTemplate(PulseTemplate, ParameterConstrainer, MeasurementDefiner):
    """A sequence of different PulseTemplates.
    
    SequencePulseTemplate allows to group several
    PulseTemplates (subtemplates) into one larger sequence,
    i.e., when instantiating a pulse from a SequencePulseTemplate
    all pulses instantiated from the subtemplates are queued for
    execution right after one another.
    SequencePulseTemplate requires to specify a mapping of
    parameter declarations from its subtemplates to its own, enabling
    renaming and mathematical transformation of parameters.
    """

    def __init__(self,
                 *subtemplates: Union[PulseTemplate, MappingTuple],
                 external_parameters: Optional[Union[Iterable[str], Set[str]]]=None,
                 identifier: Optional[str]=None,
                 parameter_constraints: Optional[List[Union[str, Expression]]]=None,
                 measurements: Optional[List[MeasurementDeclaration]]=None,
                 registry: PulseRegistryType=None) -> None:
        """Create a new SequencePulseTemplate instance.

        Requires a (correctly ordered) list of subtemplates in the form
        (PulseTemplate, Dict(str -> str)) where the dictionary is a mapping between the external
        parameters exposed by this SequencePulseTemplate to the parameters declared by the
        subtemplates, specifying how the latter are derived from the former, i.e., the mapping is
        subtemplate_parameter_name -> mapping_expression (as str) where the free variables in the
        mapping_expression are parameters declared by this SequencePulseTemplate.

        The following requirements must be satisfied:
            - for each parameter declared by a subtemplate, a mapping expression must be provided
            - each free variable in a mapping expression must be declared as an external parameter
                of this SequencePulseTemplate

        Args:
            subtemplates (List(Subtemplate)): The list of subtemplates of this
                SequencePulseTemplate as tuples of the form (PulseTemplate, Dict(str -> str)).
            external_parameters (List(str)): A set of names for external parameters of this
                SequencePulseTemplate. Deprecated.
            identifier (str): A unique identifier for use in serialization. (optional)
        """
        PulseTemplate.__init__(self, identifier=identifier)
        ParameterConstrainer.__init__(self, parameter_constraints=parameter_constraints)
        MeasurementDefiner.__init__(self, measurements=measurements)

        self.__subtemplates = [MappingPulseTemplate.from_tuple(st) if isinstance(st, tuple) else st
                               for st in subtemplates]

        # check that all subtemplates live on the same channels
        defined_channels = self.__subtemplates[0].defined_channels
        for subtemplate in self.__subtemplates[1:]:
            if subtemplate.defined_channels != defined_channels:
                raise ValueError('The subtemplates are defined for different channels')

        if external_parameters:
            warnings.warn("external_parameters is an obsolete argument and will be removed in the future.",
                          category=DeprecationWarning)

        self._register(registry=registry)

    @classmethod
    def concatenate(cls, *pulse_templates: Union[PulseTemplate, MappingTuple], **kwargs) -> 'SequencePulseTemplate':
        """Sequences the given pulse templates by creating a SequencePulseTemplate. Pulse templates that are
        SequencePulseTemplates and do not carry additional information (identifier, measurements, parameter constraints)
        are not used directly but their sub templates are.
        Args:
            *pulse_templates: Pulse templates to concatenate
            **kwargs: Forwarded to the __init__

        Returns: Concatenated templates
        """
        parsed = []
        for pt in pulse_templates:
            if (isinstance(pt, SequencePulseTemplate)
                    and pt.identifier is None
                    and not pt.measurement_declarations
                    and not pt.parameter_constraints):
                parsed.extend(pt.subtemplates)
            else:
                parsed.append(pt)
        return cls(*parsed, **kwargs)

    @property
    def parameter_names(self) -> Set[str]:
        return self.constrained_parameters.union(self.measurement_parameters, *(st.parameter_names for st in self.__subtemplates))

    @property
    def subtemplates(self) -> List[MappingPulseTemplate]:
        return self.__subtemplates

    @property
    def is_interruptable(self) -> bool:
        return any(st.is_interruptable for st in self.subtemplates)

    @cached_property
    def duration(self) -> Expression:
        return sum(sub.duration for sub in self.__subtemplates)

    @property
    def defined_channels(self) -> Set[ChannelID]:
        return self.__subtemplates[0].defined_channels if self.__subtemplates else set()

    @property
    def measurement_names(self) -> Set[str]:
        return set.union(MeasurementDefiner.measurement_names.fget(self),
                         *(st.measurement_names for st in self.subtemplates))

    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, 'Condition']) -> bool:
        """Returns the stop requirement of the first subtemplate. If a later subtemplate requires a stop the
        SequencePulseTemplate can be partially sequenced."""
        return self.__subtemplates[0].requires_stop(parameters, conditions) if self.__subtemplates else False

    def build_waveform(self,
                       parameters: Dict[str, Real],
                       channel_mapping: Dict[ChannelID, ChannelID]) -> SequenceWaveform:
        self.validate_parameter_constraints(parameters=parameters)
        return SequenceWaveform([sub_template.build_waveform(parameters,
                                                             channel_mapping=channel_mapping)
                                 for sub_template in self.__subtemplates])

    def build_sequence(self,
                       sequencer: Sequencer,
                       parameters: Dict[str, Parameter],
                       conditions: Dict[str, Condition],
                       measurement_mapping: Dict[str, str],
                       channel_mapping: Dict['ChannelID', 'ChannelID'],
                       instruction_block: InstructionBlock) -> None:
        self.validate_parameter_constraints(parameters=parameters)
        self.insert_measurement_instruction(instruction_block=instruction_block,
                                            parameters=parameters,
                                            measurement_mapping=measurement_mapping)
        for subtemplate in reversed(self.subtemplates):
            sequencer.push(subtemplate,
                           parameters=parameters,
                           conditions=conditions,
                           window_mapping=measurement_mapping,
                           channel_mapping=channel_mapping,
                           target_block=instruction_block)

    def _internal_create_program(self, *,
                                 parameters: Dict[str, Parameter],
                                 measurement_mapping: Dict[str, Optional[str]],
                                 channel_mapping: Dict[ChannelID, Optional[ChannelID]],
                                 global_transformation: Optional['Transformation'],
                                 to_single_waveform: Set[Union[str, 'PulseTemplate']],
                                 parent_loop: Loop) -> None:
        self.validate_parameter_constraints(parameters=parameters)

        try:
            measurement_parameters = {parameter_name: parameters[parameter_name].get_value()
                                      for parameter_name in self.measurement_parameters}
            duration_parameters = {parameter_name: parameters[parameter_name].get_value()
                                   for parameter_name in self.duration.variables}
        except KeyError as e:
            raise ParameterNotProvidedException(e) from e

        if self.duration.evaluate_numeric(**duration_parameters) > 0:
            measurements = self.get_measurement_windows(measurement_parameters, measurement_mapping)
            if measurements:
                parent_loop.add_measurements(measurements)

            for subtemplate in self.subtemplates:
                subtemplate._create_program(parameters=parameters,
                                            measurement_mapping=measurement_mapping,
                                            channel_mapping=channel_mapping,
                                            global_transformation=global_transformation,
                                            to_single_waveform=to_single_waveform,
                                            parent_loop=parent_loop)

    def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
        data = super().get_serialization_data(serializer)
        data['subtemplates'] = self.subtemplates

        if serializer: # compatibility to old serialization routines, deprecated
            data = dict()
            data['subtemplates'] = [serializer.dictify(subtemplate) for subtemplate in self.subtemplates]

        if self.parameter_constraints:
            data['parameter_constraints'] = [str(c) for c in self.parameter_constraints]
        if self.measurement_declarations:
            data['measurements'] = self.measurement_declarations

        return data

    @classmethod
    def deserialize(cls,
                    serializer: Optional[Serializer]=None,  # compatibility to old serialization routines, deprecated
                    **kwargs) -> 'SequencePulseTemplate':
        subtemplates = kwargs['subtemplates']
        del kwargs['subtemplates']

        if serializer: # compatibility to old serialization routines, deprecated
            subtemplates = [serializer.deserialize(st) for st in subtemplates]

        return cls(*subtemplates, **kwargs)

    @property
    def defined_channels(self) -> Set[ChannelID]:
        return self.__subtemplates[0].defined_channels

    @property
    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
        expressions = {channel: 0 for channel in self.defined_channels}

        def add_dicts(x, y):
            return {k: x[k] + y[k] for k in x}

        return functools.reduce(add_dicts, [sub.integral for sub in self.__subtemplates], expressions)





"""This module provides sequencing functionality: It defines classes and algorithms required
to translate PulseTemplates into a hardware understandable abstract instruction sequence of
instantiated pulses or waveforms.

Classes:
    - SequencingElement: Interface for objects that can be translated into instruction sequences.
    - Sequencer: Controller of the sequencing/translation process.
"""

from abc import ABCMeta, abstractmethod
from typing import Tuple, Dict, Union, Optional

# solve circular dependence of type hints
from . import conditions

from qupulse.utils.types import ChannelID
from qupulse._program.instructions import InstructionBlock, ImmutableInstructionBlock
from qupulse._program.waveforms import Waveform
from qupulse.pulses.parameters import Parameter, ConstantParameter


__all__ = ["SequencingElement", "Sequencer"]


class SequencingElement(metaclass=ABCMeta):
    """An entity which can be sequenced using Sequencer.

    DEPRECATED (September 2018). This is outdated code. For obtaining pulses for execution from pulse templates, refer
    to :meth:`.PulseTemplate.create_program`.

    See also:
        Sequencer
    """

    def __init__(self) -> None:
        pass

    def atomicity(self) -> bool:
        """Is the element translated to a single EXECInstruction with one waveform"""
        raise NotImplementedError()

    @abstractmethod
    def build_sequence(self,
                       sequencer: 'Sequencer',
                       parameters: Dict[str, Parameter],
                       conditions: Dict[str, 'conditions.Condition'],
                       measurement_mapping: Dict[str, Optional[str]],
                       channel_mapping: Dict[ChannelID, Optional[ChannelID]],
                       instruction_block: InstructionBlock) -> None:
        """Translate this SequencingElement into an instruction sequence for the given
        instruction_block using sequencer and the given parameter and condition sets.

        Implementation guide: Use instruction_block methods to add instructions or create new InstructionBlocks. Use
        sequencer to push child elements to the translation stack.

        Args:
            sequencer: The Sequencer object coordinating the current sequencing process.
            parameters: A mapping of parameter names to Parameter objects.
            conditions: A mapping of condition identifiers to Condition
            measurement_mapping: A mapping of measurement window names. Windows that are mapped to None are omitted.
            channel_mapping: A mapping of channel names. Channels that are mapped to None are omitted.
            instruction_block: The instruction block into which instructions resulting from the translation of this
                SequencingElement will be placed.
        """

    @abstractmethod
    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, 'conditions.Condition']) -> bool:
        """Return True if this SequencingElement cannot be translated yet.

        Sequencer will check requires_stop() before calling build_sequence(). If requires_stop()
        returns True, Sequencer interrupts the current translation process and will not call
        build_sequence().

        Implementation guide: requires_stop() should only return True, if this SequencingElement
        cannot be translated, i.e., the return value should only depend on the parameters/conditions
        of this SequencingElement, not on possible child elements.
        If this SequencingElement contains a child element which requires a stop, it should be
        pushed to the sequencing stack nonetheless. The requires_stop information of the child
        will be regarded during translation of that element.

        Args:
            parameters: A mapping of parameter names to Parameter objects.
            conditions: A mapping of condition identifiers to Condition objects.
        Returns:
            True, if this SequencingElement cannot be translated yet. False, otherwise.
        """


class Sequencer:
    """Translates tree structures of SequencingElement objects to linear instruction sequences.

    DEPRECATED (September 2018). This is outdated code. For obtaining pulses for execution from pulse templates, refer
    to :meth:`.PulseTemplate.create_program`.

    The concept of the sequencing process itself is described in detail in Section 1.4 of the
    documentation. Sequencer controls the process and invokes the translation functions of
    SequencingElements on the sequencing stack, which in turn implement the details of the
    translation of that specific SequencingElement.

    Sequencer manages a main InstructionBlock into which all instructions are translated by default.
    Since additional InstructionBlocks may be created during the sequencing process due to looping
    and branching, Sequencer maintains several several sequencing stacks - one for each block -
    simultaneously and continues the translation until no stack holds objects that may be
    translated anymore before compiling the final sequence and interrupting or finishing the
    sequencing process.

    See also:
        SequencingElement
    """

    StackElement = Tuple[SequencingElement, Dict[str, Parameter], Dict[str, 'conditions.Condition'], Dict[str,str]]

    def __init__(self) -> None:
        """Create a Sequencer."""
        super().__init__()
        self.__waveforms = dict()  # type: Dict[int, Waveform]
        self.__main_block = InstructionBlock()
        self.__sequencing_stacks = \
            {self.__main_block: []}  # type: Dict[InstructionBlock, List[Sequencer.StackElement]]

    def push(self,
             sequencing_element: SequencingElement,
             parameters: Optional[Dict[str, Union[Parameter, float]]]=None,
             conditions: Optional[Dict[str, 'conditions.Condition']]=None,
             *,
             window_mapping: Optional[Dict[str, Optional[str]]]=None,
             channel_mapping: Optional[Dict[ChannelID, Optional[ChannelID]]]=None,
             target_block: Optional[InstructionBlock]=None) -> None:
        """Add an element to the translation stack of the target_block with the given set of
         parameters.

        The element will be on top of the stack, i.e., it is the first to be translated if no
        subsequent calls to push with the same target_block occur.

        Args:
            sequencing_element (SequencingElement): The SequencingElement to push to the stack.
            parameters (Dict(str -> (Parameter or float)): A mapping of parameters names defined
                in the SequencingElement to Parameter objects or constant float values. In the
                latter case, the float values are encapsulated into ConstantParameter objects.
                Optional, if no conditions are defined by the SequencingElement. (default: None)
            conditions (Dict(str -> Condition)): A mapping of condition identifier defined by the
                SequencingElement to Condition objects. Optional, if no conditions are defined by
                the SequencingElement. (default: None)
            window_mapping (Dict(str -> str)): Mapping of the measurement window names of the sequence element
            channel_mapping (Dict(ChannelID -> ChannelID)): Mapping of the defined channels
            target_block (InstructionBlock): The instruction block into which instructions resulting
                from the translation of the SequencingElement will be placed. Optional. If not
                provided, the main instruction block will be targeted. (default: None)
        """
        if parameters is None:
            parameters = dict()
        if conditions is None:
            conditions = dict()
        if target_block is None:
            target_block = self.__main_block
        if window_mapping is None:
            if hasattr(sequencing_element, 'measurement_names'):
                window_mapping = {wn: wn for wn in sequencing_element.measurement_names}
            else:
                window_mapping = dict()
        if channel_mapping is None:
            if hasattr(sequencing_element, 'defined_channels'):
                channel_mapping = {cn: cn for cn in sequencing_element.defined_channels}
            else:
                channel_mapping = dict()
        for (key, value) in parameters.items():
            if not isinstance(value, Parameter):
                parameters[key] = ConstantParameter(value)

        if target_block not in self.__sequencing_stacks:
            self.__sequencing_stacks[target_block] = []

        self.__sequencing_stacks[target_block].append((sequencing_element, parameters, conditions, window_mapping,
                                                       channel_mapping))

    def build(self) -> ImmutableInstructionBlock:
        """Start the translation process. Translate all elements currently on the translation stacks
        into an InstructionBlock hierarchy.

        Processes all sequencing stacks (for each InstructionBlock) until each stack is either
        empty or its topmost element requires a stop. If build is called after a previous
        translation process where some elements required a stop (i.e., has_finished returned False),
        it will append new instructions to the previously generated and returned blocks.

        Returns:
            The instruction block (hierarchy) resulting from the translation of the (remaining)
                SequencingElements on the sequencing stacks.
        """
        if not self.has_finished():
            shall_continue = True # shall_continue will only be False, if the first element on all
                                  # stacks requires a stop or all stacks are empty
            while shall_continue:
                shall_continue = False
                for target_block, sequencing_stack in self.__sequencing_stacks.copy().items():
                    while sequencing_stack:
                        (element, parameters, conditions, window_mapping, channel_mapping) = sequencing_stack[-1]
                        if not element.requires_stop(parameters, conditions):
                            shall_continue |= True
                            sequencing_stack.pop()
                            element.build_sequence(self, parameters, conditions, window_mapping,
                                                   channel_mapping, target_block)
                        else: break

        return ImmutableInstructionBlock(self.__main_block, dict())

    def has_finished(self) -> bool:
        """Check whether all translation stacks are empty. Indicates that the translation is
        complete.

        Note that has_finished will return False, if there are stack elements that require a stop.
        In this case, calling build will only have an effect if these elements no longer require a
        stop, e.g. when required measurement results have been acquired since the last translation.

        Returns:
            Returns True, if all translation stacks are empty, i.e., the translation is complete.
        """
        return not any(self.__sequencing_stacks.values())



""" This module provides serialization and storage functionality.

Classes:
    - StorageBackend: Abstract representation of a data storage.
    - FilesystemBackend: Implementation of a file system data storage.
    - ZipFileBackend: Like FilesystemBackend but inside a single zip file instead of a directory
    - CachingBackend: A caching decorator for StorageBackends.
    - Serializable: An interface for serializable objects.
    - PulseStorage: High-level management object for loading and storing and transparently (de)serializing serializable objects.

Deprecated Classes:
    - Serializer: Converts Serializables to a serial representation as a string and vice-versa.

Functions:
    - get_default_pulse_registry: Returns the default pulse registry
    - set_default_pulse_registry: Set the default pulse registry
    - new_default_pulse_registry: Reset the default pulse registry with an empty mapping
    - convert_stored_pulse_in_storage: Converts a single Serializable stored using the deprecated Serializer class format into the PulseStorage format.
    - convert_pulses_in_storage: Converts all Serializables stored in a StorageBackend using the deprecated Serializer class format into the PulseStorage format.
"""

from abc import ABCMeta, abstractmethod
from typing import Dict, Any, Optional, NamedTuple, Union, Mapping, MutableMapping, Set, Callable, Iterator, Iterable
import os
import zipfile
import tempfile
import json
import weakref
import warnings
import gc
import importlib
import warnings
from contextlib import contextmanager

from qupulse.utils.types import DocStringABCMeta

__all__ = ["StorageBackend", "FilesystemBackend", "ZipFileBackend", "CachingBackend", "Serializable", "Serializer",
           "AnonymousSerializable", "DictBackend", "PulseStorage",
           "convert_pulses_in_storage", "convert_stored_pulse_in_storage", "PulseRegistryType", "get_default_pulse_registry",
           "set_default_pulse_registry", "new_default_pulse_registry"]


class StorageBackend(metaclass=ABCMeta):
    """A backend to store data/files in.

    Used as an abstraction of file systems/databases for the serializer.
    """

    @abstractmethod
    def put(self, identifier: str, data: str, overwrite: bool=False) -> None:
        """Stores the data string identified by identifier.

        Args:
            identifier (str): A unique identifier/name for the data to be stored.
            data (str): A serialized string of data to be stored.
            overwrite (bool): Set to True, if already existing data shall be overwritten.
                (default: False)
        Raises:
            FileExistsError if overwrite is False and there already exists data which
                is associated with the given identifier.
        """

    def __setitem__(self, identifier: str, data: str) -> None:
        self.put(identifier, data)

    @abstractmethod
    def get(self, identifier: str) -> str:
        """Retrieves the data string with the given identifier.

        Args:
            identifier (str): The identifier of the data to be retrieved.
        Returns:
            A serialized string of the data associated with the given identifier, if present.
        Raises:
            KeyError if no data is associated with the given identifier.
        """

    def __getitem__(self, identifier: str) -> str:
        return self.get(identifier)

    @abstractmethod
    def exists(self, identifier: str) -> bool:
        """Checks if data is stored for the given identifier.

        Args:
            identifier (str): The identifier for which presence of data shall be checked.
        Returns:
            True, if stored data is associated with the given identifier.
        """

    def __contains__(self, identifier: str) -> bool:
        return self.exists(identifier)

    @abstractmethod
    def delete(self, identifier: str) -> None:
        """Deletes data of the given identifier.

        Args:
            identifier: identifier of the data to be deleted

        Raises:
            KeyError if there is no data associated with the identifier
        """

    def __delitem__(self, identifier: str) -> None:
        self.delete(identifier)

    def list_contents(self) -> Iterable[str]:
        """Returns a listing of all available identifiers.

        DEPRECATED (2018-09-20): Use property contents instead.

        Returns:
            List of all available identifiers.
        """
        warnings.warn("list_contents is deprecated. Use the property contents instead", DeprecationWarning)
        return self.contents

    @property
    def contents(self) -> Iterable[str]:
        """The identifiers of all Serializables currently stored in this StorageBackend."""
        return set(iter(self))

    @abstractmethod
    def __iter__(self) -> Iterator[str]:
        """Iterator over all identifiers of Serializables currently stored in this StorageBackend."""
        pass

    def __len__(self) -> int:
        return len(self.contents)


class FilesystemBackend(StorageBackend):
    """A StorageBackend implementation based on a regular filesystem.

    Data will be stored in plain text files in a directory. The directory is given in the
    constructor of this FilesystemBackend. For each data item, a separate file is created an named
    after the corresponding identifier. If the directory does not exist, it is not created unless the create_if_missing
    argument is explicitly set to True.
    """

    def __init__(self, root: str='.', create_if_missing: bool=False) -> None:
        """Creates a new FilesystemBackend.

        Args:
            root: The path of the directory in which all data files are located. (default: ".",
                i.e. the current directory)
            create_if_missing: If False, do not create the specified directory if it does not exist. (default: False)
        Raises:
            NotADirectoryError: if root is not a valid directory path.
        """
        if not os.path.exists(root) and create_if_missing:
            os.makedirs(root)
        if not os.path.isdir(root):
            raise NotADirectoryError()
        self._root = os.path.abspath(root)

    def _path(self, identifier) -> str:
        return os.path.join(self._root, identifier + '.json')

    def put(self, identifier: str, data: str, overwrite: bool=False) -> None:
        if self.exists(identifier) and not overwrite:
            raise FileExistsError(identifier)
        path = self._path(identifier)
        with open(path, 'w') as file:
            file.write(data)

    def get(self, identifier: str) -> str:
        path = self._path(identifier)
        try:
            with open(path) as file:
                return file.read()
        except FileNotFoundError as fnf:
            raise KeyError(identifier) from fnf

    def exists(self, identifier: str) -> bool:
        path = self._path(identifier)
        return os.path.isfile(path)

    def delete(self, identifier):
        try:
            os.remove(self._path(identifier))
        except FileNotFoundError as fnf:
            raise KeyError(identifier) from fnf

    def __iter__(self) -> Iterator[str]:
        for dirpath, dirs, files in os.walk(self._root):
            return (filename for filename, ext in (os.path.splitext(file) for file in files) if ext == '.json')


class ZipFileBackend(StorageBackend):
    """A StorageBackend implementation based on a single zip file.

    Data will be stored in plain text files inside a zip file. The zip file is given
    in the constructor of this FilesystemBackend. For each data item, a separate
    file is created and named after the corresponding identifier.

    ZipFileBackend uses significantly less storage space and is faster on
    network devices, but takes longer to update because every write causes a
    complete recompression (it's not too bad)."""

    def __init__(self, root: str='./storage.zip', compression_method: int=zipfile.ZIP_DEFLATED) -> None:
        """Creates a new FilesystemBackend.

        Args:
            root: The path of the zip file in which all data files are stored. (default: "./storage.zip",
                i.e. the current directory)
            compression_method: The compression method/algorithm used to compress data in the zipfile. Accepts
                all values handled by the zipfile module (ZIP_STORED, ZIP_DEFLATED, ZIP_BZIP2, ZIP_LZMA). Please refer
                to the `zipfile docs <https://docs.python.org/3/library/zipfile.html#zipfile.ZIP_STORED>` for more
                information. (default: zipfile.ZIP_DEFLATED)
        Raises:
            NotADirectoryError if root is not a valid path.
        """
        parent, fname = os.path.split(root)
        if not os.path.isfile(root):
            if not os.path.isdir(parent):
                raise NotADirectoryError(
                    "Cannot create a ZipStorageBackend. The parent path {} is not valid.".format(parent)
                )
            z = zipfile.ZipFile(root, "w")
            z.close()
        elif not zipfile.is_zipfile(root):
            raise FileExistsError("Cannot open a ZipStorageBackend. The file {} is not a zip archive.".format(root))
        self._root = root
        self._compression_method = compression_method

    def _path(self, identifier) -> str:
        return os.path.join(identifier + '.json')

    def put(self, identifier: str, data: str, overwrite: bool=False) -> None:
        if not self.exists(identifier):
            with zipfile.ZipFile(self._root, mode='a', compression=self._compression_method) as myzip:
                path = self._path(identifier)
                myzip.writestr(path, data)
        else:
            if overwrite:
                self._update(self._path(identifier), data)
            else:
                raise FileExistsError(identifier)

    def get(self, identifier: str) -> str:
        path = self._path(identifier)
        try:
            with zipfile.ZipFile(self._root) as myzip:
                with myzip.open(path) as file:
                    return file.read().decode()
        except FileNotFoundError as fnf:
            raise KeyError(identifier) from fnf

    def exists(self, identifier: str) -> bool:
        path = self._path(identifier)
        with zipfile.ZipFile(self._root, 'r') as myzip:
            return path in myzip.namelist()

    def delete(self, identifier: str) -> None:
        if not self.exists(identifier):
            raise KeyError(identifier)
        self._update(self._path(identifier), None)

    def _update(self, filename: str, data: Optional[str]) -> None:
        # generate a temp file
        tmpfd, tmpname = tempfile.mkstemp(dir=os.path.dirname(self._root))
        os.close(tmpfd)

        # create a temp copy of the archive without filename            
        with zipfile.ZipFile(self._root, 'r') as zin:
            with zipfile.ZipFile(tmpname, 'w') as zout:
                zout.comment = zin.comment # preserve the comment
                for item in zin.infolist():
                    if item.filename != filename:
                        zout.writestr(item, zin.read(item.filename))

        # replace with the temp archive
        os.remove(self._root)
        os.rename(tmpname, self._root)

        # now add filename with its new data
        if data is not None:
            with zipfile.ZipFile(self._root, mode='a', compression=self._compression_method) as zf:
                zf.writestr(filename, data)

    def __iter__(self) -> Iterator[str]:
        with zipfile.ZipFile(self._root, 'r') as myzip:
            return (filename
                    for filename, ext in (os.path.splitext(file) for file in myzip.namelist())
                    if ext == '.json')


class CachingBackend(StorageBackend):
    """Adds naive memory caching functionality to another StorageBackend.

    CachingBackend relies on another StorageBackend to provide real data IO functionality which
    it extends by caching already opened files in memory for faster subsequent access.

    Note that it does not automatically clear the cache at any time and thus will consume increasing amounts of memory
    over time. Use the :meth:`clear_cache` method to clear the cache manually.

    DEPRECATED (2018-09-20): PulseStorage now already provides chaching around StorageBackends, rendering CachingBackend
    obsolete.
    """

    def __init__(self, backend: StorageBackend) -> None:
        """Creates a new CachingBackend.

        Args:
            backend (StorageBackend): A StorageBackend that provides data
                IO functionality.
        """
        warnings.warn("CachingBackend is obsolete due to PulseStorage already offering caching functionality.",
                      DeprecationWarning)
        self._backend = backend
        self._cache = {}

    def put(self, identifier: str, data: str, overwrite: bool=False) -> None:
        if identifier in self._cache and not overwrite:
            raise FileExistsError(identifier)
        self._backend.put(identifier, data, overwrite)
        self._cache[identifier] = data

    def get(self, identifier: str) -> str:
        if identifier not in self._cache:
            self._cache[identifier] = self._backend.get(identifier)
        return self._cache[identifier]

    def exists(self, identifier: str) -> bool:
        return self._backend.exists(identifier)

    def delete(self, identifier: str) -> None:
        self._backend.delete(identifier)
        if identifier in self._cache:
            del self._cache[identifier]

    def __iter__(self) -> Iterator[str]:
        return iter(self._backend)

    def clear_cache(self) -> None:
        self._cache = dict()


class DictBackend(StorageBackend):
    """DictBackend uses a dictionary to store Serializables in memory.

    Doing so, it does not provide any persistent storage functionality.
    """
    def __init__(self) -> None:
        self._cache = {}

    def put(self, identifier: str, data: str, overwrite: bool=False) -> None:
        if identifier in self._cache and not overwrite:
            raise FileExistsError(identifier)
        self._cache[identifier] = data

    def get(self, identifier: str) -> str:
        return self._cache[identifier]

    def exists(self, identifier: str) -> bool:
        return identifier in self._cache
    
    @property
    def storage(self) -> Dict[str, str]:
        return self._cache

    def delete(self, identifier: str) -> None:
        del self._cache[identifier]

    def __iter__(self) -> Iterator[str]:
        return iter(self._cache)


class DeserializationCallbackFinder:
    def __init__(self):
        self._storage = {}
        self.auto_import = True
        self.qctoolkit_alias = True

    def __setitem__(self, type_name: str, callback: Callable):
        self._storage[type_name] = callback

    def __getitem__(self, type_name: str) -> Callable:
        if self.qctoolkit_alias and type_name.startswith('qctoolkit.'):
            type_name = type_name.replace('qctoolkit.', 'qupulse.')

        if self.auto_import and type_name not in self._storage:
            module_name = '.'.join(type_name.split('.')[:-1])
            importlib.import_module(module_name)

        return self._storage[type_name]

    def __contains__(self, type_name) -> bool:
        return type_name in self._storage


class SerializableMeta(DocStringABCMeta):
    deserialization_callbacks = DeserializationCallbackFinder()

    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)

        type_identifier = getattr(cls, 'get_type_identifier')()

        try:
            deserialization_function = getattr(cls, 'deserialize')
        except AttributeError:
            deserialization_function = cls
        mcs.deserialization_callbacks[type_identifier] = deserialization_function

        return cls


PulseRegistryType = Optional[MutableMapping[str, 'Serializable']]
default_pulse_registry = None # type: PulseRegistryType


def get_default_pulse_registry() -> PulseRegistryType:
    """Returns the current default pulse registry."""
    return default_pulse_registry


def set_default_pulse_registry(new_default_registry: PulseRegistryType) -> None:
    """Sets the default pulse registry.

    Args:
        new_default_registry: Any PulseRegistryType  object (i.e., mutable mapping) which will become the new default
            pulse registry.
    """
    global default_pulse_registry
    default_pulse_registry = new_default_registry


def new_default_pulse_registry() -> None:
    """Sets a new empty default pulse registry.

    The new registry is a newly created weakref.WeakValueDictionry().
    """
    set_default_pulse_registry(weakref.WeakValueDictionary())


class Serializable(metaclass=SerializableMeta):
    """Any object that can be converted into a serialized representation for storage and back.

    Serializable is the interface used by PulseStorage to obtain representations of objects that
    need to be stored. It essentially provides the methods get_serialization_data, which returns
    a dictionary which contains all relevant properties of the Serializable object encoded as
    basic Python types, and deserialize, which is able to reconstruct the object from given
    such a dictionary.

    Additionally, a Serializable object MAY have a unique identifier, which indicates towards
    the PulseStorage that this object should be stored as a separate data item and accessed by
    reference instead of possibly embedding it into a containing Serializable's representation.

    All Serializables MUST automatically register themselves with the default pulse registry on
    construction unless an explicit other registry is provided to them as construction argument.
    This MUST be implemented by all subclasses of Serializable by calling `Serializable._register` at some point
    in their __init__ method.
    This is intended to prevent accidental duplicate usage of identifiers by failing early.

    See also:
        PulseStorage
    """

    type_identifier_name = '#type'
    identifier_name = '#identifier'

    def __init__(self, identifier: Optional[str]=None) -> None:
        """Initializes a Serializable.

        Args:
            identifier: An optional, non-empty identifier for this Serializable.
                If set, this Serializable will always be stored as a separate data item and never
                be embedded.
        Raises:
            ValueError: If identifier is the empty string
        """
        super().__init__()

        if identifier == '':
            raise ValueError("Identifier must not be empty.")
        self.__identifier = identifier

    def _register(self, registry: Optional[PulseRegistryType]=None) -> None:
        """Registers the Serializable in the global registry.

        This method MUST be called by subclasses at some point during init.
        Args:
            registry: An optional mutable mapping where the Serializable is registered. If None, it gets registered in
                the default_pulse_registry.
        Raises:
            RuntimeError: If a Serializable with the same name is already registered.
        """
        if registry is None:
            registry = default_pulse_registry

        if self.identifier and registry is not None:
            if self.identifier in registry and isinstance(registry, weakref.WeakValueDictionary):
                # trigger garbage collection in case the registered object isn't referenced anymore
                gc.collect(2)

            if self.identifier in registry:
                raise RuntimeError('Pulse with name already exists', self.identifier)

            registry[self.identifier] = self

    @property
    def identifier(self) -> Optional[str]:
        """The (optional) identifier of this Serializable. Either a non-empty string or None."""
        return self.__identifier

    def get_serialization_data(self, serializer: Optional['Serializer']=None) -> Dict[str, Any]:
        """Returns all data relevant for serialization as a dictionary containing only base types.

        Implementation hint:
        In the old serialization routines, if the Serializable contains complex objects which are itself
        Serializables, a serialized representation for these MUST be obtained by calling the dictify()
        method of serializer. The reason is that serializer may decide to either return a dictionary
        to embed or only a reference to the Serializable subelement. This is DEPRECATED behavior as of May 2018.
        In the new routines, this will happen automatically and every Serializable is only responsible for
        returning it's own data and leave nested Serializables in object form.

        For the transition time where both implementations are
        available, implementations of this method should support the old and new routines, using
        the presence of the serializer argument to differentiate between both. Don't make use of
        the implementation in this base class when implementing this method for the old routines.

        Args:
            serializer (Serializer): DEPRECATED (May 2018).A Serializer instance used to serialize
                complex subelements of this Serializable.
        Returns:
            A dictionary of Python base types (strings, integers, lists/tuples containing these,
                etc..) which fully represent the relevant properties of this Serializable for
                storing and later reconstruction as a Python object.
        """
        if serializer:
            warnings.warn("{c}.get_serialization_data(*) was called with a serializer argument, indicating deprecated behavior. Please switch to the new serialization routines.".format(c=self.__class__.__name__), DeprecationWarning)

        if self.identifier:
            return {self.type_identifier_name: self.get_type_identifier(), self.identifier_name: self.identifier}
        else:
            return {self.type_identifier_name: self.get_type_identifier()}

    @classmethod
    def get_type_identifier(cls) -> str:
        return "{}.{}".format(cls.__module__, cls.__name__)

    @classmethod
    def deserialize(cls, serializer: Optional['Serializer']=None, **kwargs) -> 'Serializable':
        """Reconstructs the Serializable object from a dictionary.

        Implementation hint:
        For greater clarity, implementations of this method should be precise in their return value,
        i.e., give their exact class name, and also replace the kwargs argument by a list of
        arguments required, i.e., those returned by get_serialization_data.
        Using old serialization routines, if this Serializable contains complex objects which are itself
        of type Serializable, their dictionary representations MUST be converted into objects using
        serializers deserialize() method. This is DEPRECATED behavior.
        Using the new routines, a serializable is only responsible to decode it's own dictionary,
        not those of nested objects (i.e., all incoming arguments are already processed by the
        serialization routines).
        For the transition time where both variants are
        available, implementations of this method should support the old and new routines, using
        the presence of the serializer argument to differentiate between both. For the new routines,
        just call this base class function.
        After the transition period, subclasses likely need not implement deserialize separately anymore at all.

        Args:
            serializer: DEPRECATED (May 2018). A serializer instance used when deserializing subelements.
            **kwargs: All relevant properties of the object as keyword arguments. For every (key,value)
                pair returned by get_serialization_data, the same pair is given as keyword argument as input
                to this method.
         """
        if serializer:
            warnings.warn("{c}.deserialize(*) was called with a serializer argument, indicating deprecated behavior. Please switch to the new serialization routines.".format(c=cls.__name__), DeprecationWarning)

        return cls(**kwargs)

    def renamed(self, new_identifier: str, registry: Optional[PulseRegistryType]=None) -> 'Serializable':
        """Returns a copy of the Serializable with its identifier set to new_identifier.

        Args:
            new_identifier: The identifier of the new copy of this Serializable.
            registry: The pulse registry the copy of this Serializable will register in. If None, the default pulse
                registry will be used. Optional.
        """
        data = self.get_serialization_data()
        data.pop(Serializable.type_identifier_name)
        data.pop(Serializable.identifier_name, None)
        return self.deserialize(registry=registry, identifier=new_identifier, **data)


class AnonymousSerializable:
    """Any object that can be converted into a serialized representation for storage and back which NEVER has an
    identifier. This class is used for implicit serialization and does not work necessarily with dicts.

    The type information is not saved explicitly but implicitly by the position in the JSON-document.

    See also:
        Serializable

    # todo (lumip, 2018-05-30): this does not really have a purpose, especially in the new serialization ecosystem.. we should deprecate and remove it
    """

    def get_serialization_data(self) -> Any:
        """Return all data relevant for serialization as a JSON compatible type that is accepted as constructor argument

        Returns:
            A JSON compatible type that can be used to construct an equal object.
        """
        raise NotImplementedError()


class Serializer(object):
    """Serializes Serializable objects and stores them persistently.

    DEPRECATED as of May 2018. Serializer will be superseeded by the new serialization routines and
    PulseStorage class.

    Serializer provides methods to enable the conversion of Serializable objects (including nested
    Serializables) into (nested) dictionaries and serialized JSON-encodings of these and vice-versa.
    Additionally, it can also store these representations persistently using a StorageBackend
    instance.

    See also:
        Serializable
    """

    __FileEntry = NamedTuple("FileEntry", [('serialization', str), ('serializable', Serializable)])

    def __init__(self, storage_backend: StorageBackend) -> None:
        """Creates a Serializer.

        Args:
            storage_backend (StorageBackend): The StorageBackend all objects will be stored in.
        """
        self.__subpulses = dict() # type: Dict[str, Serializer.__FileEntry]
        self.__storage_backend = storage_backend

        warnings.warn("Serializer is deprecated. Please switch to the new serialization routines.", DeprecationWarning)

    def dictify(self, serializable: Serializable) -> Union[str, Dict[str, Any]]:
        """Converts a Serializable into a dictionary representation.

        The Serializable is converted by calling its get_serialization_data() method. If it contains
        nested Serializables, these are also converted into dictionarys (or references), yielding
        a single dictionary representation of the outermost Serializable where all nested
        Serializables are either completely embedded or referenced by identifier.

        Args:
            serializable (Serializabe): The Serializable object to convert.
        Returns:
            A serialization dictionary, i.e., a dictionary of Python base types (strings, integers,
                lists/tuples containing these, etc..) which fully represent the relevant properties
                of the given Serializable for storing and later reconstruction as a Python object.
                Nested Serializables are either embedded or referenced by identifier.
        Raises:
            Exception if an identifier is assigned twice to different Serializable objects
                encountered by this Serializer during the conversion.
        See also:
            Serializable.get_serialization_data
        """
        repr_ = serializable.get_serialization_data(serializer=self)
        repr_['type'] = self.get_type_identifier(serializable)
        identifier = serializable.identifier
        if identifier is None:
            return repr_
        else:
            if identifier in self.__subpulses:
                if self.__subpulses[identifier].serializable is not serializable:
                    raise Exception("Identifier '{}' assigned twice.".format(identifier))
            else:
                self.__subpulses[identifier] = Serializer.__FileEntry(repr_, serializable)
            return identifier

    def __collect_dictionaries(self, serializable: Serializable) -> Dict[str, Dict[str, Any]]:
        """Converts a Serializable into a collection of dictionary representations.

        The Serializable is converted by calling its get_serialization_data() method. If it contains
        nested Serializables, these are also converted into dictionarys (or references), yielding
        a dictionary representation of the outermost Serializable where all nested
        Serializables are either completely embedded or referenced by identifier as it is returned
        by dictify. If nested Serializables shall be stored separately, their dictionary
        representations are collected. Collection_dictionaries returns a dictionary of all
        serialization dictionaries where the keys are the identifiers of the Serializables.

        Args:
            serializable (Serializabe): The Serializable object to convert.
        Returns:
            A dictionary containing serialization dictionary for each separately stored Serializable
                nested in the given Serializable.
        See also:
            dictify
        """
        self.__subpulses = dict()
        repr_ = self.dictify(serializable)
        filedict = dict()
        for identifier in self.__subpulses:
            filedict[identifier] = self.__subpulses[identifier].serialization
        if isinstance(repr_, dict):
            filedict[''] = repr_
        return filedict

    @staticmethod
    def get_type_identifier(obj: Any) -> str:
        """Returns a unique type identifier for any object.

        Args:
            obj: The object for which to obtain a type identifier.
        Returns:
            The type identifier as a string.
        """
        return "{}.{}".format(obj.__module__, obj.__class__.__name__)

    def serialize(self, serializable: Serializable, overwrite=False) -> None:
        """Serializes and stores a Serializable.

        The given Serializable and all nested Serializables that are to be stored separately will be
        converted into a serial string representation by obtaining their dictionary representation,
        encoding them as a JSON-string and storing them in the StorageBackend.

        If no identifier is given for the Serializable, "main" will be used.

        If an identifier is already in use in the StorageBackend, associated data will be replaced.

        Args:
            serializable (Serializable): The Serializable to serialize and store
        """
        warnings.warn("Serializer is deprecated. Please switch to the new serialization routines.", DeprecationWarning)
        repr_ = self.__collect_dictionaries(serializable)
        for identifier in repr_:
            storage_identifier = identifier
            if identifier == '':
                storage_identifier = 'main'
            json_str = json.dumps(repr_[identifier], indent=4, sort_keys=True, cls=ExtendedJSONEncoder)
            self.__storage_backend.put(storage_identifier, json_str, overwrite)

    def deserialize(self, representation: Union[str, Dict[str, Any]]) -> Serializable:
        """Loads a stored Serializable object or converts a dictionary representation back to the
            corresponding Serializable.

        Args:
            representation: A serialization dictionary representing a Serializable object or the
                identifier of a Serializable object to load from the StorageBackend.
        Returns:
            The Serializable object instantiated from its serialized representation.
        See also:
            Serializable.deserialize
        """
        warnings.warn("Serializer is deprecated. Please switch to the new serialization routines.", DeprecationWarning)
        if isinstance(representation, str):
            if representation in self.__subpulses:
                return self.__subpulses[representation].serializable

        if isinstance(representation, str):
            repr_ = json.loads(self.__storage_backend.get(representation))
            repr_['identifier'] = representation
        else:
            repr_ = dict(representation)

        module_name, class_name = repr_['type'].rsplit('.', 1)
        module = __import__(module_name, fromlist=[class_name])
        class_ = getattr(module, class_name)

        repr_to_store = repr_.copy()
        repr_.pop('type')

        serializable = class_.deserialize(self, **repr_)

        if 'identifier' in repr_:
            identifier = repr_['identifier']
            self.__subpulses[identifier] = self.__FileEntry(repr_, serializable)
        return serializable


class PulseStorage(MutableMapping[str, Serializable]):
    """The central storage management for pulses.

    Provides a dictionary interface for loading and storing pulses based on any StorageBackend implementation.
    Takes care of serialization and deserialization of pulses/Serializables in the process. Every Serializable
    with an identifier will be stored as a separate entity, even if it is nested in another Serializable that is stored.
    Serializables containing named Serializables will just store a reference to those which will be transparently
    resolved by PulseStorage during loading.

    PulseStorage employs caching, i.e., once a Serializable/pulse is loaded, it will be kept in memory and subsequent
    fetches will be instantaneous. At the same time, all changes to Serializables known to PulseStorage will immediately
    be flushed to the storage backend.

    Note that it is currently not possible to store a Serializable under a different identifier than the one it holds,
    i.e. you cannot store a Serializable `serializable` with identifier 'foo' under identifier 'bar' by calling
    `pulse_storage['bar'] = serializable`. This will currently result in a ValueError.

    It is also not possible to overwrite a Serializable using the dictionary interface. To explicitly overwrite a
    Serializable in the storage, use the `overwrite` method.

    PulseStorage can be used as the default pulse registry.
    All Serializables (and thus pulses) will automatically register themselves with the default pulse registry on
    construction unless an explicit other registry is provided to them as construction argument.
    This is intended to prevent accidental duplicate usage of identifiers by failing early. Setting
    a PulseStorage as pulse default registry also implies that all created Serializables are automatically stored
    in the storage backend.
    See Also:
        PulseStorage.set_to_default_registry
        PulseStorage.as_default_registry
    """
    StorageEntry = NamedTuple('StorageEntry', [('serialization', str), ('serializable', Serializable)])

    def __init__(self,
                 storage_backend: StorageBackend) -> None:
        """Create a PulseStorage instance.

        Args:
            storage_backend: The StorageBackend representing the permanent storage of the PulseStorage. Serializables
                are stored to and read from here.
        """
        self._storage_backend = storage_backend

        self._temporary_storage = dict() # type: Dict[str, StorageEntry]
        self._transaction_storage = None

    def _deserialize(self, serialization: str) -> Serializable:
        decoder = JSONSerializableDecoder(storage=self)
        serializable = decoder.decode(serialization)
        return serializable

    def _load_and_deserialize(self, identifier: str) -> StorageEntry:
        serialization = self._storage_backend[identifier]
        serializable = self._deserialize(serialization)
        self._temporary_storage[identifier] = PulseStorage.StorageEntry(serialization=serialization,
                                                                        serializable=serializable)
        return self._temporary_storage[identifier]

    @property
    def temporary_storage(self) -> Dict[str, StorageEntry]:
        """The in-memory temporary storage.

        Contains all Serializables that have been loaded during the lifetime of this PulseStorage object."""
        return self._temporary_storage

    def __contains__(self, identifier) -> bool:
        return identifier in self._temporary_storage or identifier in self._storage_backend

    def __getitem__(self, identifier: str) -> Serializable:
        """Fetch a Serializable.

        If the Serializable is not present in temporary storage, it will be loaded and deserialized from the storage
        backend.

        Args:
            identifier: The identifier of the Serializable to load.
        """
        if identifier not in self._temporary_storage:
            self._load_and_deserialize(identifier)
        return self._temporary_storage[identifier].serializable

    def __setitem__(self, identifier: str, serializable: Serializable) -> None:
        """Store a Serializable in the PulseStorage.

        Note that it is currently not possible to store a Serializable under a different identifier than the one it holds,
        i.e. you cannot store a Serializable `serializable` with identifier 'foo' under identifier 'bar' by calling
        `pulse_storage['bar'] = serializable`. This will currently result in a ValueError.

        It is also not possible to overwrite a Serializable using the dictionary interface. To explicitly overwrite a
        Serializable in the storage, use the `overwrite` method.

        Args:
            identifier: The identifier to store the Serializable under. Has to be identical to `serialziable.identifier`.
            serializable: The Serializable object to be stored.
        Raises:
            ValueError: if the given identifier argument does not match the identifier of the serializable
        """
        if identifier != serializable.identifier: # address issue #272: https://github.com/qutech/qupulse/issues/272
            raise ValueError("Storing a Serializable under a different than its own internal identifier is currently"
                             " not supported! If you want to rename the serializable, please use the "
                             "Serializable.renamed() method to obtain a renamed copy which can then be stored with "
                             "the new identifier.\n"
                             "If you think that storing under a different identifier without explicit renaming should"
                             "a supported feature, please contribute to our ongoing discussion about this on:\n"
                             "https://github.com/qutech/qupulse/issues/272")
        if identifier in self._temporary_storage:
            if self.temporary_storage[identifier].serializable is serializable:
                return
            else:
                raise RuntimeError('Identifier assigned twice with different objects', identifier)
        elif identifier in self._storage_backend:
            raise RuntimeError('Identifier already assigned in storage backend', identifier)
        self.overwrite(identifier, serializable)

    def __delitem__(self, identifier: str) -> None:
        """Delete an item from temporary storage and storage backend.

        Does not raise an error if the deleted pulse is only in the storage backend. Assumes that all pulses
        contained in temporary storage are always also contained in the storage backend.
        Args:
            identifier: Identifier of the Serializable to delete
        """
        del self._storage_backend[identifier]
        try:
            del self._temporary_storage[identifier]
        except KeyError:
            pass

    @property
    def contents(self) -> Iterable[str]:
        return self._storage_backend.list_contents()

    def __len__(self) -> int:
        return len(self._storage_backend)

    def __iter__(self) -> Iterator[str]:
        return iter(self._storage_backend)

    def overwrite(self, identifier: str, serializable: Serializable) -> None:
        """Explicitly overwrites a pulse.

        Calling this method will overwrite the entity currently stored under the given identifier by the
        provided serializable. It does _not_ overwrite nested Serializable objects contained in serializable. If you
        want to overwrite those as well, do that explicitely.

        Args:
              identifier: The identifier to store serializable under.
            serializable: The Serializable object to be stored.
        """

        is_transaction_begin = (self._transaction_storage is None)
        try:
            if is_transaction_begin:
                self._transaction_storage = dict()

            encoder = JSONSerializableEncoder(self, sort_keys=True, indent=4)

            serialization_data = serializable.get_serialization_data()
            serialized = encoder.encode(serialization_data)
            self._transaction_storage[identifier] = self.StorageEntry(serialized, serializable)

            if is_transaction_begin:
                for identifier, entry in self._transaction_storage.items():
                    self._storage_backend.put(identifier, entry.serialization, overwrite=True)
                self._temporary_storage.update(**self._transaction_storage)

        finally:
            if is_transaction_begin:
                self._transaction_storage = None

    def clear(self) -> None:
        """Clears the temporary storage.

        Does not affect the storage backend."""
        self._temporary_storage.clear()

    @contextmanager
    def as_default_registry(self) -> Any:
        """Returns context manager to use this PulseStorage as the default pulse registry only within a with-statement."""
        global default_pulse_registry
        previous_registry = default_pulse_registry
        default_pulse_registry = self
        try:
            yield self
        finally:
            default_pulse_registry = previous_registry

    def set_to_default_registry(self) -> None:
        """Promotes this PulseStorage object to be the default pulse registry.

        All Serializables (and thus pulses) will automatically register themselves with the default pulse registry on
        construction unless an explicit other registry is provided to them as construction argument.
        This is intended to prevent accidental duplicate usage of identifiers by failing early. Setting
        a PulseStorage as pulse default registry also implies that all created Serializables are automatically stored
        in the storage backend."""
        global default_pulse_registry
        default_pulse_registry = self


class JSONSerializableDecoder(json.JSONDecoder):
    """JSONDecoder for Serializables.

    Automatically follows references to nested Serializables during deserializing."""

    def __init__(self, storage: Mapping, *args, **kwargs) -> None:
        """Creates a new JSONSerialzableDecoder object.

        Args:
            storage: Any mapping of identifier to Serializable objects. Will be used to resolve references to nested
                Serializables. Usually a PulseStorage object.
            *args: Any other positional argument will be passed on to JSONDecoder constructor.
            **kwargs: Any keyword argument will be passed on to JSONDecoder.
        See Also:
            JSONDecoder
        """
        super().__init__(*args, object_hook=self.filter_serializables, **kwargs)

        self.storage = storage

    def filter_serializables(self, obj_dict) -> Any:
        if Serializable.type_identifier_name in obj_dict:
            type_identifier = obj_dict.pop(Serializable.type_identifier_name)

            if Serializable.identifier_name in obj_dict:
                obj_identifier = obj_dict.pop(Serializable.identifier_name)
            else:
                obj_identifier = None

            if type_identifier == 'reference':
                if not obj_identifier:
                    raise RuntimeError('Reference without identifier')
                return self.storage[obj_identifier]

            else:
                deserialization_callback = SerializableMeta.deserialization_callbacks[type_identifier]

                # if the storage is the default registry, we would get conflicts when the Serializable tries to register
                # itself on construction. Pass an empty dict as registry keyword argument in this case.
                # calling PulseStorage objects will take care of registering.
                # (solution to issue #301: https://github.com/qutech/qupulse/issues/301 )
                registry = None
                if get_default_pulse_registry() is self.storage:
                    registry = dict()

                return deserialization_callback(identifier=obj_identifier, registry=registry, **obj_dict)
        return obj_dict


class JSONSerializableEncoder(json.JSONEncoder):
    """JSONEncoder for Serializables.

    Ensures that nested Serializables are stored as separate entities and embedded in the parent Serializable's
    serialization by reference."""

    def __init__(self, storage: MutableMapping, *args, **kwargs) -> None:
        """Creates a new JSONSerialzableDecoder object.

            Args:
                storage: Any mapping of identifier to Serializable objects. Will be used to store nested
                    Serializables. Usually a PulseStorage object.
                *args: Any other positional argument will be passed on to JSONEncoder constructor.
                **kwargs: Any keyword argument will be passed on to JSONEncoder.
            See Also:
                JSONEncoder
        """
        super().__init__(*args, **kwargs)

        self.storage = storage

    def default(self, o: Any) -> Any:
        if isinstance(o, Serializable):
            if o.identifier:
                if o.identifier not in self.storage:
                    self.storage[o.identifier] = o
                elif o is not self.storage[o.identifier]:
                    raise RuntimeError('Trying to store a subpulse with an identifier that is already taken.')


                return {Serializable.type_identifier_name: 'reference',
                        Serializable.identifier_name: o.identifier}
            else:
                return o.get_serialization_data()

        elif isinstance(o, AnonymousSerializable):
            return o.get_serialization_data()

        elif type(o) is set:
            return list(o)

        else:
            return super().default(o)


class ExtendedJSONEncoder(json.JSONEncoder):
    """Encodes AnonymousSerializable and sets as lists.

    Used by Serializer.

    Deprecated as of May 2018. To be replaced by JSONSerializableEncoder."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def default(self, o: Any) -> Any:
        if isinstance(o, AnonymousSerializable):
            return o.get_serialization_data()
        elif type(o) is set:
            return list(o)
        else:
            return super().default(o)


def convert_stored_pulse_in_storage(identifier: str, source_storage: StorageBackend, dest_storage: StorageBackend) -> None:
    """Converts a pulse from the old to the new serialization format.

    The pulse with the given identifier is completely (including subpulses) converted from the old serialization format
    read from a given source storage to the new serialization format and written to a given destination storage.

    Args:
        identifier (str): The identifier of the pulse to convert.
        source_storage (StorageBackend): A StorageBackend containing the pulse identified by the identifier argument in the old serialization format.
        dest_storage (StorageBackend): A StorageBackend the converted pulse will be written to in the new serialization format.
    Raises:
        ValueError: if the dest_storage StorageBackend contains identifiers also assigned in source_storage.
    """
    if dest_storage.list_contents().intersection(source_storage.list_contents()):
        raise ValueError("dest_storage already contains pulses with the same ids. Aborting to prevent inconsistencies for duplicate keys.")
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', DeprecationWarning)
        serializer = Serializer(source_storage)
        pulse_storage = PulseStorage(dest_storage)
        serializable = serializer.deserialize(identifier)
        pulse_storage.overwrite(identifier, serializable)


def convert_pulses_in_storage(source_storage: StorageBackend, dest_storage: StorageBackend) -> None:
    """Converts all pulses from the old to the new serialization format.

        All pulses in a given source storage are completely (including subpulses) converted from the old serialization format
        to the new serialization format and written to a given destination storage.

        Args:
            source_storage (StorageBackend): A StorageBackend containing pulses in the old serialization format.
            dest_storage (StorageBackend): A StorageBackend the converted pulses will be written to in the new serialization format.
        Raises:
            ValueError: if the dest_storage StorageBackend contains identifiers also assigned in source_storage.
        """
    if dest_storage.list_contents().intersection(source_storage.list_contents()):
        raise ValueError("dest_storage already contains pulses with the same ids. Aborting to prevent inconsistencies for duplicate keys.")
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', DeprecationWarning)
        serializer = Serializer(source_storage)
        pulse_storage = PulseStorage(dest_storage)
        for identifier in source_storage.list_contents():
            serializable = serializer.deserialize(identifier)
            pulse_storage.overwrite(identifier, serializable)



from qupulse.pulses.mapping_pulse_template import MappingPulseTemplate

__all__ = ["MappingPulseTemplate"]

import warnings
warnings.warn("MappingPulseTemplate was moved from qupulse.pulses.pulse_template_parameter_mapping to "
              "qupulse.pulses.mapping_pulse_template. Please consider fixing your stored pulse templates by loading "
              "and storing them anew.", DeprecationWarning)

from qupulse.serialization import SerializableMeta
SerializableMeta.deserialization_callbacks["qupulse.pulses.pulse_template_parameter_mapping.MappingPulseTemplate"] = SerializableMeta.deserialization_callbacks[MappingPulseTemplate.get_type_identifier()]



from typing import Optional, List, Union, Set, Dict, Sequence, Any
from numbers import Real
import itertools
import numbers

import numpy as np

from qupulse.utils.types import ChannelID
from qupulse.expressions import Expression, ExpressionScalar
from qupulse.pulses.conditions import Condition
from qupulse._program.waveforms import TableWaveform, TableWaveformEntry
from qupulse.pulses.parameters import Parameter, ParameterNotProvidedException, ParameterConstraint,\
    ParameterConstrainer
from qupulse.pulses.pulse_template import AtomicPulseTemplate, MeasurementDeclaration
from qupulse.pulses.table_pulse_template import TableEntry, EntryInInit
from qupulse.pulses.multi_channel_pulse_template import MultiChannelWaveform
from qupulse.serialization import Serializer, PulseRegistryType


__all__ = ["PointWaveform", "PointPulseTemplate", "PointPulseEntry", "PointWaveformEntry", "InvalidPointDimension"]


PointWaveform = TableWaveform
PointWaveformEntry = TableWaveformEntry


class PointPulseEntry(TableEntry):
    def instantiate(self, parameters: Dict[str, numbers.Real], num_channels: int) -> Sequence[PointWaveformEntry]:
        t = self.t.evaluate_numeric(**parameters)
        vs = self.v.evaluate_numeric(**parameters)

        if isinstance(vs, numbers.Number):
            vs = np.full(num_channels, vs, dtype=type(vs))
        elif len(vs) != num_channels:
            raise InvalidPointDimension(expected=num_channels, received=len(vs))

        return tuple(PointWaveformEntry(t, v, self.interp)
                     for v in vs)


class PointPulseTemplate(AtomicPulseTemplate, ParameterConstrainer):
    def __init__(self,
                 time_point_tuple_list: List[EntryInInit],
                 channel_names: Sequence[ChannelID],
                 *,
                 parameter_constraints: Optional[List[Union[str, ParameterConstraint]]]=None,
                 measurements: Optional[List[MeasurementDeclaration]]=None,
                 identifier: Optional[str]=None,
                 registry: PulseRegistryType=None) -> None:

        AtomicPulseTemplate.__init__(self, identifier=identifier, measurements=measurements)
        ParameterConstrainer.__init__(self, parameter_constraints=parameter_constraints)

        self._channels = tuple(channel_names)
        self._entries = [PointPulseEntry(*tpt)
                         for tpt in time_point_tuple_list]

        self._register(registry=registry)

    @property
    def defined_channels(self) -> Set[ChannelID]:
        return set(self._channels)

    def build_waveform(self,
                       parameters: Dict[str, Real],
                       channel_mapping: Dict[ChannelID, Optional[ChannelID]]) -> Optional[TableWaveform]:
        self.validate_parameter_constraints(parameters)

        if all(channel_mapping[channel] is None
               for channel in self.defined_channels):
            return None

        if self.duration.evaluate_numeric(**parameters) == 0:
            return None

        mapped_channels = tuple(channel_mapping[c] for c in self._channels)

        waveform_entries = list([] for _ in range(len(self._channels)))
        for entry in self._entries:
            instantiated_entries = entry.instantiate(parameters, len(self._channels))
            for ch_entries, wf_entry in zip(waveform_entries, instantiated_entries):
                ch_entries.append(wf_entry)

        if waveform_entries[0][0].t > 0:
            for ch_entries in waveform_entries:
                ch_entries[:0] = [PointWaveformEntry(0, ch_entries[0].v, ch_entries[0].interp)]

        # filter mappings to None
        channel_entries = [(ch, ch_entries)
                           for (ch, ch_entries) in zip(mapped_channels, waveform_entries)
                           if ch is not None]
        mapped_channels, waveform_entries = zip(*channel_entries)

        waveforms = [PointWaveform(mapped_channel, ch_entries)
                     for mapped_channel, ch_entries in zip(mapped_channels, waveform_entries)]

        if len(waveforms) == 1:
            return waveforms.pop()
        else:
            return MultiChannelWaveform(waveforms)

    @property
    def point_pulse_entries(self) -> Sequence[PointPulseEntry]:
        return self._entries

    def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
        data = super().get_serialization_data(serializer)

        if serializer: # compatibility to old serialization routines, deprecated
            data = dict()

        data['time_point_tuple_list'] = [entry.get_serialization_data() for entry in self._entries]
        data['channel_names'] = self._channels
        if self.parameter_constraints:
            data['parameter_constraints'] = [str(c) for c in self.parameter_constraints]
        if self.measurement_declarations:
            data['measurements'] = self.measurement_declarations
        return data

    @property
    def duration(self) -> Expression:
        return self._entries[-1].t

    @property
    def point_parameters(self) -> Set[str]:
        return set(
            var
            for time, point, *_ in self._entries
            for var in itertools.chain(time.variables, point.variables)
        )

    @property
    def parameter_names(self) -> Set[str]:
        return self.point_parameters | self.measurement_parameters | self.constrained_parameters

    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, Condition]) -> bool:
        try:
            return any(
                parameters[name].requires_stop
                for name in self.parameter_names
            )
        except KeyError as key_error:
            raise ParameterNotProvidedException(str(key_error)) from key_error

    @property
    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
        expressions = {channel: 0 for channel in self._channels}
        for first_entry, second_entry in zip(self._entries[:-1], self._entries[1:]):
            substitutions = {'t0': ExpressionScalar(first_entry.t).sympified_expression,
                             't1': ExpressionScalar(second_entry.t).sympified_expression}

            for i, channel in enumerate(self._channels):
                substitutions['v0'] = ExpressionScalar(first_entry.v[i]).sympified_expression
                substitutions['v1'] = ExpressionScalar(second_entry.v[i]).sympified_expression
                expressions[channel] += first_entry.interp.integral.sympified_expression.subs(substitutions)

        expressions = {c: ExpressionScalar(expressions[c]) for c in expressions}
        return expressions


class InvalidPointDimension(Exception):
    def __init__(self, expected, received):
        super().__init__('Expected a point of dimension {} but received {}'.format(expected, received))
        self.expected = expected
        self.received = received



"""This is the central package for defining pulses. All :class:`~qupulse.pulses.pulse_template.PulseTemplate`
subclasses that are final and ready to be used are imported here with their recommended abbreviation as an alias."""

from qupulse.pulses.function_pulse_template import FunctionPulseTemplate as FunctionPT
from qupulse.pulses.loop_pulse_template import ForLoopPulseTemplate as ForLoopPT
from qupulse.pulses.multi_channel_pulse_template import AtomicMultiChannelPulseTemplate as AtomicMultiChannelPT
from qupulse.pulses.mapping_pulse_template import MappingPulseTemplate as MappingPT
from qupulse.pulses.repetition_pulse_template import RepetitionPulseTemplate as RepetitionPT
from qupulse.pulses.sequence_pulse_template import SequencePulseTemplate as SequencePT
from qupulse.pulses.table_pulse_template import TablePulseTemplate as TablePT
from qupulse.pulses.point_pulse_template import PointPulseTemplate as PointPT

import warnings
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    # ensure this is included.. it adds a deserialization handler for pulse_template_parameter_mapping.MappingPT which is not present otherwise
    import qupulse.pulses.pulse_template_parameter_mapping

__all__ = ["FunctionPT", "ForLoopPT", "AtomicMultiChannelPT", "MappingPT", "RepetitionPT", "SequencePT", "TablePT",
           "PointPT"]




"""This module defines LoopPulseTemplate, a higher-order hierarchical pulse template that loops
another PulseTemplate based on a condition."""


from typing import Dict, Set, Optional, Any, Union, Tuple, Generator, Sequence, cast
import warnings

import sympy
from cached_property import cached_property

from qupulse.serialization import Serializer, PulseRegistryType
from qupulse._program._loop import Loop

from qupulse.expressions import ExpressionScalar
from qupulse.utils import checked_int_cast
from qupulse.pulses.parameters import Parameter, ConstantParameter, InvalidParameterNameException, ParameterConstrainer, ParameterNotProvidedException
from qupulse.pulses.pulse_template import PulseTemplate, ChannelID
from qupulse.pulses.conditions import Condition, ConditionMissingException
from qupulse._program.instructions import InstructionBlock
from qupulse.pulses.sequencing import Sequencer
from qupulse._program.waveforms import SequenceWaveform as ForLoopWaveform
from qupulse.pulses.measurement import MeasurementDefiner, MeasurementDeclaration

__all__ = ['ForLoopPulseTemplate', 'LoopPulseTemplate', 'LoopIndexNotUsedException']


class LoopPulseTemplate(PulseTemplate):
    """Base class for loop based pulse templates. This class is still abstract and cannot be instantiated."""
    def __init__(self, body: PulseTemplate,
                 identifier: Optional[str]):
        super().__init__(identifier=identifier)
        self.__body = body

    @property
    def body(self) -> PulseTemplate:
        return self.__body

    @property
    def defined_channels(self) -> Set['ChannelID']:
        return self.__body.defined_channels

    @property
    def measurement_names(self) -> Set[str]:
        return self.__body.measurement_names

    @property
    def is_interruptable(self):
        raise NotImplementedError()  # pragma: no cover


class ParametrizedRange:
    """Like the builtin python range but with parameters."""
    def __init__(self, *args, **kwargs):
        """Positional and keyword arguments cannot be mixed.

        Args:
            *args: Interpreted as ``(start, )`` or ``(start, stop[, step])``
            **kwargs: Expected to contain ``start``, ``stop`` and ``step``
        Raises:
            TypeError: If positional and keyword arguments are mixed
            KeyError: If keyword arguments but one of ``start``, ``stop`` or ``step`` is missing
        """
        if args and kwargs:
            raise TypeError('ParametrizedRange only takes either positional or keyword arguments')
        elif kwargs:
            start = kwargs['start']
            stop = kwargs['stop']
            step = kwargs['step']
        elif len(args) in (1, 2, 3):
            if len(args) == 3:
                start, stop, step = args
            elif len(args) == 2:
                (start, stop), step = args, 1
            elif len(args) == 1:
                start, (stop,), step = 0, args, 1
        else:
            raise TypeError('ParametrizedRange expected 1 to 3 arguments, got {}'.format(len(args)))

        self.start = ExpressionScalar.make(start)
        self.stop = ExpressionScalar.make(stop)
        self.step = ExpressionScalar.make(step)

    def to_tuple(self) -> Tuple[Any, Any, Any]:
        """Return a simple representation of the range which is useful for comparison and serialization"""
        return (self.start.get_serialization_data(),
                self.stop.get_serialization_data(),
                self.step.get_serialization_data())

    def to_range(self, parameters: Dict[str, Any]) -> range:
        return range(checked_int_cast(self.start.evaluate_numeric(**parameters)),
                     checked_int_cast(self.stop.evaluate_numeric(**parameters)),
                     checked_int_cast(self.step.evaluate_numeric(**parameters)))

    @property
    def parameter_names(self) -> Set[str]:
        return set(self.start.variables) | set(self.stop.variables) | set(self.step.variables)


class ForLoopPulseTemplate(LoopPulseTemplate, MeasurementDefiner, ParameterConstrainer):
    """This pulse template allows looping through an parametrized integer range and provides the loop index as a
    parameter to the body. If you do not need the index in the pulse template, consider using
    :class:`~qupulse.pulses.repetition_pulse_template.RepetitionPulseTemplate`"""
    def __init__(self,
                 body: PulseTemplate,
                 loop_index: str,
                 loop_range: Union[int,
                                   range,
                                   str,
                                   Tuple[Any, Any],
                                   Tuple[Any, Any, Any],
                                   ParametrizedRange],
                 identifier: Optional[str]=None,
                 *,
                 measurements: Optional[Sequence[MeasurementDeclaration]]=None,
                 parameter_constraints: Optional[Sequence]=None,
                 registry: PulseRegistryType=None) -> None:
        """
        Args:
            body: The loop body. It is expected to have `loop_index` as an parameter
            loop_index: Loop index of the for loop
            loop_range: Range to loop through
            identifier: Used for serialization
        """
        LoopPulseTemplate.__init__(self, body=body, identifier=identifier)
        MeasurementDefiner.__init__(self, measurements=measurements)
        ParameterConstrainer.__init__(self, parameter_constraints=parameter_constraints)

        if isinstance(loop_range, ParametrizedRange):
            self._loop_range = loop_range
        elif isinstance(loop_range, (int, str)):
            self._loop_range = ParametrizedRange(loop_range)
        elif isinstance(loop_range, (tuple, list)):
            self._loop_range = ParametrizedRange(*loop_range)
        elif isinstance(loop_range, range):
            self._loop_range = ParametrizedRange(start=loop_range.start,
                                                 stop=loop_range.stop,
                                                 step=loop_range.step)
        else:
            raise ValueError('loop_range is not valid')

        if not loop_index.isidentifier():
            raise InvalidParameterNameException(loop_index)
        body_parameters = self.body.parameter_names
        if loop_index not in body_parameters:
            raise LoopIndexNotUsedException(loop_index, body_parameters)
        self._loop_index = loop_index

        if self.loop_index in self.constrained_parameters:
            constraints = [str(constraint) for constraint in self.parameter_constraints
                           if self._loop_index in constraint.affected_parameters]
            warnings.warn("ForLoopPulseTemplate was created with a constraint on a variable shadowing the loop index.\n" \
                          "This will not constrain the actual loop index but introduce a new parameter.\n" \
                          "To constrain the loop index, put the constraint in the body subtemplate.\n" \
                          "Loop index is {} and offending constraints are: {}".format(self._loop_index, constraints))

        self._register(registry=registry)

    @property
    def loop_index(self) -> str:
        return self._loop_index

    @property
    def loop_range(self) -> ParametrizedRange:
        return self._loop_range

    @property
    def measurement_names(self) -> Set[str]:
        return LoopPulseTemplate.measurement_names.fget(self) | MeasurementDefiner.measurement_names.fget(self)

    @cached_property
    def duration(self) -> ExpressionScalar:
        step_size = self._loop_range.step.sympified_expression
        loop_index = sympy.symbols(self._loop_index)
        sum_index = sympy.symbols(self._loop_index)

        # replace loop_index with sum_index dependable expression
        body_duration = self.body.duration.sympified_expression.subs({loop_index: self._loop_range.start.sympified_expression + sum_index*step_size})

        # number of sum contributions
        step_count = sympy.ceiling((self._loop_range.stop.sympified_expression-self._loop_range.start.sympified_expression) / step_size)
        sum_start = 0
        sum_stop = sum_start + (sympy.functions.Max(step_count, 1) - 1)

        # expression used if step_count >= 0
        finite_duration_expression = sympy.Sum(body_duration, (sum_index, sum_start, sum_stop))

        duration_expression = sympy.Piecewise((0, step_count <= 0),
                                              (finite_duration_expression, True))

        return ExpressionScalar(duration_expression)

    @property
    def parameter_names(self) -> Set[str]:
        parameter_names = self.body.parameter_names.copy()
        parameter_names.remove(self._loop_index)
        return parameter_names | self._loop_range.parameter_names | self.constrained_parameters | self.measurement_parameters

    def _body_parameter_generator(self, parameters: Dict[str, Parameter], forward=True) -> Generator:
        loop_range_parameters = dict((parameter_name, parameters[parameter_name].get_value())
                                     for parameter_name in self._loop_range.parameter_names)
        loop_range = self._loop_range.to_range(loop_range_parameters)

        parameters = dict((parameter_name, parameters[parameter_name])
                          for parameter_name in self.body.parameter_names if parameter_name != self._loop_index)
        loop_range = loop_range if forward else reversed(loop_range)
        for loop_index_value in loop_range:
            local_parameters = parameters.copy()
            local_parameters[self._loop_index] = ConstantParameter(loop_index_value)
            yield local_parameters

    def build_sequence(self,
                       sequencer: Sequencer,
                       parameters: Dict[str, Parameter],
                       conditions: Dict[str, Condition],
                       measurement_mapping: Dict[str, str],
                       channel_mapping: Dict[ChannelID, ChannelID],
                       instruction_block: InstructionBlock) -> None:
        self.validate_parameter_constraints(parameters=parameters)

        self.insert_measurement_instruction(instruction_block=instruction_block,
                                            parameters=parameters,
                                            measurement_mapping=measurement_mapping)

        for local_parameters in self._body_parameter_generator(parameters, forward=False):
            sequencer.push(self.body,
                           parameters=local_parameters,
                           conditions=conditions,
                           window_mapping=measurement_mapping,
                           channel_mapping=channel_mapping,
                           target_block=instruction_block)

    def _internal_create_program(self, *,
                                 parameters: Dict[str, Parameter],
                                 measurement_mapping: Dict[str, Optional[str]],
                                 channel_mapping: Dict[ChannelID, Optional[ChannelID]],
                                 global_transformation: Optional['Transformation'],
                                 to_single_waveform: Set[Union[str, 'PulseTemplate']],
                                 parent_loop: Loop) -> None:
        self.validate_parameter_constraints(parameters=parameters)

        try:
            measurement_parameters = {parameter_name: parameters[parameter_name].get_value()
                                      for parameter_name in self.measurement_parameters}
            duration_parameters = {parameter_name: parameters[parameter_name].get_value()
                                      for parameter_name in self.duration.variables}
        except KeyError as e:
            raise ParameterNotProvidedException(str(e)) from e

        if self.duration.evaluate_numeric(**duration_parameters) > 0:
            measurements = self.get_measurement_windows(measurement_parameters, measurement_mapping)
            if measurements:
                parent_loop.add_measurements(measurements)

            for local_parameters in self._body_parameter_generator(parameters, forward=True):
                self.body._create_program(parameters=local_parameters,
                                          measurement_mapping=measurement_mapping,
                                          channel_mapping=channel_mapping,
                                          global_transformation=global_transformation,
                                          to_single_waveform=to_single_waveform,
                                          parent_loop=parent_loop)

    def build_waveform(self, parameters: Dict[str, Parameter]) -> ForLoopWaveform:
        return ForLoopWaveform([self.body.build_waveform(local_parameters)
                                for local_parameters in self._body_parameter_generator(parameters, forward=True)])

    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, 'Condition']) -> bool:
        return any(parameters[parameter_name].requires_stop for parameter_name in self._loop_range.parameter_names)

    def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
        data = super().get_serialization_data(serializer)

        data['body'] = self.body

        if serializer: # compatibility to old serialization routines, deprecated
            data = dict()
            data['body'] = serializer.dictify(self.body)

        data['loop_range'] = self._loop_range.to_tuple()
        data['loop_index'] = self._loop_index

        if self.parameter_constraints:
            data['parameter_constraints'] = [str(c) for c in self.parameter_constraints]
        if self.measurement_declarations:
            data['measurements'] = self.measurement_declarations

        return data

    @classmethod
    def deserialize(cls, serializer: Optional[Serializer]=None, **kwargs) -> 'ForLoopPulseTemplate':
        if serializer: # compatibility to old serialization routines, deprecated
            kwargs['body'] = cast(PulseTemplate, serializer.deserialize(kwargs['body']))
        return super().deserialize(None, **kwargs)

    @property
    def integral(self) -> Dict[ChannelID, ExpressionScalar]:

        step_size = self._loop_range.step.sympified_expression
        loop_index = sympy.symbols(self._loop_index)
        sum_index = sympy.symbols(self._loop_index)

        body_integrals = self.body.integral
        body_integrals = {
            c: body_integrals[c].sympified_expression.subs(
                {loop_index: self._loop_range.start.sympified_expression + sum_index*step_size}
            )
            for c in body_integrals
        }

        # number of sum contributions
        step_count = sympy.ceiling((self._loop_range.stop.sympified_expression-self._loop_range.start.sympified_expression) / step_size)
        sum_start = 0
        sum_stop = sum_start + (sympy.functions.Max(step_count, 1) - 1)

        for c in body_integrals:
            channel_integral_expr = sympy.Sum(body_integrals[c], (sum_index, sum_start, sum_stop))
            body_integrals[c] = ExpressionScalar(channel_integral_expr)

        return body_integrals


class WhileLoopPulseTemplate(LoopPulseTemplate):
    """Conditional looping in a pulse.
    
    A LoopPulseTemplate is a PulseTemplate whose body is repeated
    during execution as long as a certain condition holds.
    """
    
    def __init__(self, condition: str,
                 body: PulseTemplate,
                 identifier: Optional[str]=None,
                 registry: PulseRegistryType=None) -> None:
        """Create a new LoopPulseTemplate instance.

        Args:
            condition (str): A unique identifier for the looping condition. Will be used to obtain
                the Condition object from the mapping passed in during the sequencing process.
            body (PulseTemplate): The PulseTemplate which will be repeated as long as the condition
                holds.
            identifier (str): A unique identifier for use in serialization. (optional)
        """
        super().__init__(body=body, identifier=identifier)
        self._condition = condition
        self._register(registry=registry)

    def __str__(self) -> str:
        return "LoopPulseTemplate: Condition <{}>, Body <{}>".format(self._condition, self.body)

    @property
    def condition(self) -> str:
        """This LoopPulseTemplate's condition."""
        return self._condition

    @property
    def parameter_names(self) -> Set[str]:
        return self.body.parameter_names

    @property
    def duration(self) -> ExpressionScalar:
        return ExpressionScalar('nan')

    def __obtain_condition_object(self, conditions: Dict[str, Condition]) -> Condition:
        try:
            return conditions[self._condition]
        except:
            raise ConditionMissingException(self._condition)

    def build_sequence(self,
                       sequencer: Sequencer,
                       parameters: Dict[str, Parameter],
                       conditions: Dict[str, Condition],
                       measurement_mapping: Dict[str, str],
                       channel_mapping: Dict[ChannelID, ChannelID],
                       instruction_block: InstructionBlock) -> None:
        self.__obtain_condition_object(conditions).build_sequence_loop(self,
                                                                       self.body,
                                                                       sequencer,
                                                                       parameters,
                                                                       conditions,
                                                                       measurement_mapping,
                                                                       channel_mapping,
                                                                       instruction_block)

    def _internal_create_program(self, *, # pragma: no cover
                                 parameters: Dict[str, Parameter],
                                 measurement_mapping: Dict[str, Optional[str]],
                                 channel_mapping: Dict[ChannelID, Optional[ChannelID]],
                                 parent_loop: Loop) -> None:
        raise NotImplementedError("create_program() does not handle conditions/triggers right now and cannot "
                                  "be meaningfully implemented for a WhileLoopPulseTemplate")

    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, Condition]) -> bool:
        return self.__obtain_condition_object(conditions).requires_stop()

    def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
        data = super().get_serialization_data(serializer)
        data['body'] = self.body

        if serializer: # compatibility to old serialization routines, deprecated
            data = dict()
            data['body'] = serializer.dictify(self.body)

        data['condition'] = self._condition

        return data

    @classmethod
    def deserialize(cls, serializer: Optional[Serializer]=None, **kwargs) -> 'WhileLoopPulseTemplate':
        if serializer: # compatibility to old serialization routines, deprecated
            kwargs['body'] = serializer.deserialize(kwargs['body'])

        return super().deserialize(**kwargs)

    @property
    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
        return {c: ExpressionScalar('nan') for c in self.body.defined_channels}


class LoopIndexNotUsedException(Exception):
    def __init__(self, loop_index: str, body_parameter_names: Set[str]):
        self.loop_index = loop_index
        self.body_parameter_names = body_parameter_names

    def __str__(self) -> str:
        return "The parameter {} is missing in the body's parameter names: {}".format(self.loop_index,
                                                                                      self.body_parameter_names)



"""This module defines MultiChannelPulseTemplate, which allows the combination of several
AtomicPulseTemplates into a single template spanning several channels.

Classes:
    - MultiChannelPulseTemplate: A pulse template defined for several channels by combining pulse
        templates
    - MultiChannelWaveform: A waveform defined for several channels by combining waveforms
"""

from typing import Dict, List, Optional, Any, Iterable, Union, Set, Sequence
import numbers
import warnings

import numpy


from qupulse.serialization import Serializer, PulseRegistryType

from qupulse.pulses.conditions import Condition
from qupulse.utils.types import ChannelID, TimeType
from qupulse._program.waveforms import MultiChannelWaveform
from qupulse.pulses.pulse_template import PulseTemplate, AtomicPulseTemplate
from qupulse.pulses.mapping_pulse_template import MappingPulseTemplate, MappingTuple
from qupulse.pulses.parameters import Parameter, ParameterConstrainer
from qupulse.pulses.measurement import MeasurementDeclaration, MeasurementWindow
from qupulse.expressions import Expression, ExpressionScalar

__all__ = ["AtomicMultiChannelPulseTemplate"]


class AtomicMultiChannelPulseTemplate(AtomicPulseTemplate, ParameterConstrainer):
    def __init__(self,
                 *subtemplates: Union[AtomicPulseTemplate, MappingTuple, MappingPulseTemplate],
                 external_parameters: Optional[Set[str]]=None,
                 identifier: Optional[str]=None,
                 parameter_constraints: Optional[List]=None,
                 measurements: Optional[List[MeasurementDeclaration]]=None,
                 registry: PulseRegistryType=None) -> None:
        AtomicPulseTemplate.__init__(self, identifier=identifier, measurements=measurements)
        ParameterConstrainer.__init__(self, parameter_constraints=parameter_constraints)

        self._subtemplates = [st if isinstance(st, PulseTemplate) else MappingPulseTemplate.from_tuple(st) for st in
                              subtemplates]

        for subtemplate in self._subtemplates:
            if isinstance(subtemplate, AtomicPulseTemplate):
                continue
            elif isinstance(subtemplate, MappingPulseTemplate):
                if isinstance(subtemplate.template, AtomicPulseTemplate):
                    continue
                else:
                    raise TypeError('Non atomic subtemplate of MappingPulseTemplate: {}'.format(subtemplate.template))
            else:
                raise TypeError('Non atomic subtemplate: {}'.format(subtemplate))

        if not self._subtemplates:
            raise ValueError('Cannot create empty MultiChannelPulseTemplate')

        defined_channels = [st.defined_channels for st in self._subtemplates]

        # check there are no intersections between channels
        for i, channels_i in enumerate(defined_channels):
            for j, channels_j in enumerate(defined_channels[i + 1:]):
                if channels_i & channels_j:
                    raise ChannelMappingException('subtemplate {}'.format(i + 1),
                                                  'subtemplate {}'.format(i + 2 + j),
                                                  (channels_i | channels_j).pop())

        if external_parameters is not None:
            warnings.warn("external_parameters is an obsolete argument and will be removed in the future.",
                          category=DeprecationWarning)

        duration = self._subtemplates[0].duration
        for subtemplate in self._subtemplates[1:]:
            if (duration == subtemplate.duration) is True:
                continue
            else:
                raise ValueError('Could not assert duration equality of {} and {}'.format(duration,
                                                                                          subtemplate.duration))

        self._register(registry=registry)

    @property
    def duration(self) -> Expression:
        return self._subtemplates[0].duration

    @property
    def parameter_names(self) -> Set[str]:
        return set.union(self.measurement_parameters,
                         self.constrained_parameters,
                         *(st.parameter_names for st in self._subtemplates))

    @property
    def subtemplates(self) -> Sequence[AtomicPulseTemplate]:
        return self._subtemplates

    @property
    def defined_channels(self) -> Set[ChannelID]:
        return set.union(*(st.defined_channels for st in self._subtemplates))

    @property
    def measurement_names(self) -> Set[str]:
        return super().measurement_names.union(*(st.measurement_names for st in self._subtemplates))

    def build_waveform(self, parameters: Dict[str, numbers.Real],
                       channel_mapping: Dict[ChannelID, Optional[ChannelID]]) -> Optional['MultiChannelWaveform']:
        self.validate_parameter_constraints(parameters=parameters)

        sub_waveforms = []
        for subtemplate in self.subtemplates:
            sub_waveform = subtemplate.build_waveform(parameters,
                                                      channel_mapping=channel_mapping)
            if sub_waveform is not None:
                sub_waveforms.append(sub_waveform)

        if len(sub_waveforms) == 0:
            return None
        if len(sub_waveforms) == 1:
            return sub_waveforms[0]
        else:
            return MultiChannelWaveform(sub_waveforms)

    def get_measurement_windows(self,
                                parameters: Dict[str, numbers.Real],
                                measurement_mapping: Dict[str, Optional[str]]) -> List[MeasurementWindow]:
        measurements = super().get_measurement_windows(parameters=parameters,
                                                       measurement_mapping=measurement_mapping)
        for st in self.subtemplates:
            measurements.extend(st.get_measurement_windows(parameters=parameters,
                                                           measurement_mapping=measurement_mapping))
        return measurements

    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, 'Condition']) -> bool:
        return any(st.requires_stop(parameters, conditions) for st in self._subtemplates)

    def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
        data = super().get_serialization_data(serializer)
        data['subtemplates'] = self.subtemplates

        if serializer: # compatibility to old serialization routines, deprecated
            data = dict()
            data['subtemplates'] = [serializer.dictify(subtemplate) for subtemplate in self.subtemplates]

        if self.parameter_constraints:
            data['parameter_constraints'] = [str(constraint) for constraint in self.parameter_constraints]
        if self.measurement_declarations:
            data['measurements'] = self.measurement_declarations

        return data

    @classmethod
    def deserialize(cls, serializer: Optional[Serializer]=None, **kwargs) -> 'AtomicMultiChannelPulseTemplate':
        subtemplates = kwargs['subtemplates']
        del kwargs['subtemplates']

        if serializer: # compatibility to old serialization routines, deprecated
            subtemplates = [serializer.deserialize(st) for st in subtemplates]

        return cls(*subtemplates, **kwargs)

    @property
    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
        expressions = dict()
        for subtemplate in self._subtemplates:
            expressions.update(subtemplate.integral)
        return expressions


class ChannelMappingException(Exception):
    def __init__(self, obj1, obj2, intersect_set):
        self.intersect_set = intersect_set
        self.obj1 = obj1
        self.obj2 = obj2

    def __str__(self) -> str:
        return 'Channel <{chs}> is defined in {o1} and {o2}'.format(chs=self.intersect_set, o1=self.obj1, o2=self.obj2)


from abc import ABCMeta, abstractmethod
from typing import Dict, Tuple

import numpy

__all__ = ['DAC']


class DAC(metaclass=ABCMeta):
    """Representation of a data acquisition card"""

    @abstractmethod
    def register_measurement_windows(self, program_name: str, windows: Dict[str, Tuple['numpy.ndarray',
                                                                                       'numpy.ndarray']]) -> None:
        """"""

    @abstractmethod
    def register_operations(self, program_name: str, operations) -> None:
        """"""

    @abstractmethod
    def arm_program(self, program_name: str) -> None:
        """"""

    @abstractmethod
    def delete_program(self, program_name) -> None:
        """"""

    @abstractmethod
    def clear(self) -> None:
        """Clears all registered programs.

        Caution: This affects all programs and waveforms on the AWG, not only those uploaded using qupulse!
        """



"""This module defines the TablePulseTemplate, one of the elementary pulse templates and its
waveform representation.

Classes:
    - TablePulseTemplate: Defines a pulse via interpolation of a sequence of (time,voltage)-pairs.
    - TableWaveform: A waveform instantiated from a TablePulseTemplate by providing values for its
        declared parameters.
"""

from typing import Union, Dict, List, Set, Optional, Any, Tuple, Sequence, NamedTuple
import numbers
import itertools
import warnings

import numpy as np
import sympy

from qupulse.utils.types import ChannelID
from qupulse.serialization import Serializer, PulseRegistryType
from qupulse.pulses.parameters import Parameter, \
    ParameterNotProvidedException, ParameterConstraint, ParameterConstrainer
from qupulse.pulses.pulse_template import AtomicPulseTemplate, MeasurementDeclaration
from qupulse.pulses.interpolation import InterpolationStrategy, LinearInterpolationStrategy, \
    HoldInterpolationStrategy, JumpInterpolationStrategy
from qupulse._program.waveforms import TableWaveform, TableWaveformEntry
from qupulse.expressions import ExpressionScalar, Expression
from qupulse.pulses.multi_channel_pulse_template import MultiChannelWaveform
from qupulse.pulses.conditions import Condition

__all__ = ["TablePulseTemplate", "concatenate"]


ValueInInit = Union[ExpressionScalar, str, numbers.Real]
EntryInInit = Union['TableEntry',
                    Tuple[ValueInInit, ValueInInit],
                    Tuple[ValueInInit, ValueInInit, Union[str, InterpolationStrategy]]]


class TableEntry(NamedTuple('TableEntry', [('t', ExpressionScalar),
                                           ('v', Expression),
                                           ('interp', InterpolationStrategy)])):
    __slots__ = ()

    def __new__(cls, t: ValueInInit, v: ValueInInit, interp: Union[str, InterpolationStrategy]='default'):
        if interp in TablePulseTemplate.interpolation_strategies:
            interp = TablePulseTemplate.interpolation_strategies[interp]
        if not isinstance(interp, InterpolationStrategy):
            raise KeyError(interp, 'is not a valid interpolation strategy')

        return super().__new__(cls, ExpressionScalar.make(t),
                                    Expression.make(v),
                                    interp)

    def instantiate(self, parameters: Dict[str, numbers.Real]) -> TableWaveformEntry:
        return TableWaveformEntry(self.t.evaluate_numeric(**parameters),
                                  self.v.evaluate_numeric(**parameters),
                                  self.interp)

    def get_serialization_data(self) -> tuple:
        return self.t.get_serialization_data(), self.v.get_serialization_data(), str(self.interp)


class TablePulseTemplate(AtomicPulseTemplate, ParameterConstrainer):
    """The TablePulseTemplate class implements pulses described by a table with time, voltage and interpolation strategy
    inputs. The interpolation strategy describes how the voltage between the entries is interpolated(see also
    InterpolationStrategy.) It can define multiple channels of which each has a separate table. If they do not have the
    same length the shorter channels are extended to the longest duration.

    If the time entries of all channels are equal it is more convenient to use the :paramrefPointPulseTemplate`."""
    interpolation_strategies = {'linear': LinearInterpolationStrategy(),
                                'hold': HoldInterpolationStrategy(),
                                'jump': JumpInterpolationStrategy(),
                                'default': HoldInterpolationStrategy()}

    def __init__(self, entries: Dict[ChannelID, Sequence[EntryInInit]],
                 identifier: Optional[str]=None,
                 *,
                 parameter_constraints: Optional[List[Union[str, ParameterConstraint]]]=None,
                 measurements: Optional[List[MeasurementDeclaration]]=None,
                 consistency_check: bool=True,
                 registry: PulseRegistryType=None) -> None:
        """
        Construct a `TablePulseTemplate` from a dict which maps channels to their entries. By default the consistency
        of the provided entries is checked. There are two static functions for convenience construction: from_array and
        from_entry_list.

        Args:
            entries: A dictionary that maps channel ids to a list of entries. An entry is a
                (time, voltage[, interpolation strategy]) tuple or a TableEntry
            identifier: Used for serialization
            parameter_constraints: Constraint list that is forwarded to the ParameterConstrainer superclass
            measurements: Measurement declaration list that is forwarded to the MeasurementDefiner superclass
            consistency_check: If True the consistency of the times will be checked on construction as far as possible
        """
        AtomicPulseTemplate.__init__(self, identifier=identifier, measurements=measurements)
        ParameterConstrainer.__init__(self, parameter_constraints=parameter_constraints)

        self._entries = dict((ch, list()) for ch in entries.keys())
        for channel, channel_entries in entries.items():
            if len(channel_entries) == 0:
                raise ValueError('Channel {} is empty'.format(channel))

            for entry in channel_entries:
                self._add_entry(channel, TableEntry(*entry))

        self._duration = self.calculate_duration()
        self._table_parameters = set(
            var
            for channel_entries in self.entries.values()
            for entry in channel_entries
            for var in itertools.chain(entry.t.variables, entry.v.variables)
        ) | self.constrained_parameters

        if self.duration == 0:
            warnings.warn('Table pulse template with duration 0 on construction.',
                          category=ZeroDurationTablePulseTemplate)

        if consistency_check:
            # perform a simple consistency check. All inequalities with more than one free variable are ignored as the
            # sympy solver does not support them

            # collect all conditions
            inequalities = [eq.sympified_expression for eq in self._parameter_constraints] +\
                           [sympy.Le(previous_entry.t.underlying_expression, entry.t.underlying_expression)
                            for channel_entries in self._entries.values()
                            for previous_entry, entry in zip(channel_entries, channel_entries[1:])]

            # test if any condition is already dissatisfied
            if any(isinstance(eq, sympy.boolalg.BooleanAtom) and bool(eq) is False
                   for eq in inequalities):
                raise ValueError('Table pulse template has impossible parametrization')

            # filter conditions that are inequalities with one free variable and test if the solution set is empty
            inequalities = [eq for eq in inequalities if isinstance(eq, sympy.Rel) and len(eq.free_symbols) == 1]
            if not sympy.reduce_inequalities(inequalities):
                raise ValueError('Table pulse template has impossible parametrization')

        self._register(registry=registry)

    def _add_entry(self, channel, new_entry: TableEntry) -> None:

        # comparisons with Expression can yield None -> use 'is True' and 'is False'
        if (new_entry.t < 0) is True:
            raise ValueError('Time parameter number {} of channel {} is negative.'.format(
                len(self._entries[channel]), channel))

        for previous_entry in self._entries[channel]:
            if (new_entry.t < previous_entry.t) is True:
                raise ValueError('Time parameter number {} of channel {} is smaller than a previous one'.format(
                    len(self._entries[channel]), channel))

        self._entries[channel].append(new_entry)

    @property
    def entries(self) -> Dict[ChannelID, List[TableEntry]]:
        return self._entries

    def get_entries_instantiated(self, parameters: Dict[str, numbers.Real]) \
            -> Dict[ChannelID, List[TableWaveformEntry]]:
        """Compute an instantiated list of the table's entries.

        Args:
            parameters (Dict(str -> Parameter)): A mapping of parameter names to Parameter objects.
        Returns:
             (float, float)-list of all table entries with concrete values provided by the given
                parameters.
        """
        if not (self.table_parameters <= set(parameters.keys())):
            raise ParameterNotProvidedException((self.table_parameters - set(parameters.keys())).pop())

        instantiated_entries = dict()  # type: Dict[ChannelID,List[TableWaveformEntry]]

        for channel, channel_entries in self._entries.items():
            instantiated = [entry.instantiate(parameters)
                            for entry in channel_entries]

            # Add (0, v) entry if wf starts at finite time
            if instantiated[0].t > 0:
                instantiated.insert(0, TableWaveformEntry(0,
                                                          instantiated[0].v,
                                                          TablePulseTemplate.interpolation_strategies['hold']))
            instantiated_entries[channel] = instantiated

        duration = max(instantiated[-1].t for instantiated in instantiated_entries.values())

        # ensure that all channels have equal duration
        for channel, instantiated in instantiated_entries.items():
            final_entry = instantiated[-1]
            if final_entry.t < duration:
                instantiated.append(TableWaveformEntry(duration,
                                                       final_entry.v,
                                                       TablePulseTemplate.interpolation_strategies['hold']))
            instantiated_entries[channel] = instantiated
        return instantiated_entries

    @property
    def table_parameters(self) -> Set[str]:
        return self._table_parameters

    @property
    def parameter_names(self) -> Set[str]:
        return self.table_parameters | self.measurement_parameters | self.constrained_parameters

    @property
    def is_interruptable(self) -> bool:
        return False

    @property
    def duration(self) -> ExpressionScalar:
        return self._duration

    def calculate_duration(self) -> ExpressionScalar:
        duration_expressions = [entries[-1].t for entries in self._entries.values()]
        duration_expression = sympy.Max(*(expr.sympified_expression for expr in duration_expressions))
        return ExpressionScalar(duration_expression)

    @property
    def defined_channels(self) -> Set[ChannelID]:
        return set(self._entries.keys())

    def requires_stop(self,
                      parameters: Dict[str, Parameter],
                      conditions: Dict[str, 'Condition']) -> bool:
        try:
            return any(
                parameters[name].requires_stop
                for name in self.parameter_names
            )
        except KeyError as key_error:
            raise ParameterNotProvidedException(str(key_error)) from key_error

    def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
        data = super().get_serialization_data(serializer)

        if serializer: # compatibility to old serialization routines, deprecated
            data = dict()

        local_data = dict(
            entries=dict(
                (channel, [entry.get_serialization_data()
                           for entry in channel_entries])
                for channel, channel_entries in self.entries.items()
            ),
            parameter_constraints=[str(c) for c in self.parameter_constraints],
            measurements=self.measurement_declarations
        )
        data.update(**local_data)
        return data

    def build_waveform(self,
                       parameters: Dict[str, numbers.Real],
                       channel_mapping: Dict[ChannelID, Optional[ChannelID]]) -> Optional[Union[TableWaveform,
                                                                                                MultiChannelWaveform]]:
        self.validate_parameter_constraints(parameters)

        if all(channel_mapping[channel] is None
               for channel in self.defined_channels):
            return None

        instantiated = [(channel_mapping[channel], instantiated_channel)
                        for channel, instantiated_channel in self.get_entries_instantiated(parameters).items()
                        if channel_mapping[channel] is not None]

        if self.duration.evaluate_numeric(**parameters) == 0:
            return None

        waveforms = [TableWaveform(*ch_instantiated)
                     for ch_instantiated in instantiated]

        if len(waveforms) == 1:
            return waveforms.pop()
        else:
            return MultiChannelWaveform(waveforms)

    @staticmethod
    def from_array(times: np.ndarray, voltages: np.ndarray, channels: List[ChannelID]) -> 'TablePulseTemplate':
        """Static constructor to build a TablePulse from numpy arrays.

        Args:
            times: 1D numpy array with time values
            voltages: 1D or 2D numpy array with voltage values
            channels: channels to define

        Returns:
            TablePulseTemplate with the given values, hold interpolation everywhere and no free
            parameters.
        """
        if times.ndim == 0 or voltages.ndim == 0:
            raise ValueError('Zero dimensional input is not accepted.')

        if times.ndim > 2 or voltages.ndim > 2:
            raise ValueError('Three or higher dimensional input is not accepted.')

        if times.ndim == 2 and times.shape[0] != len(channels):
            raise ValueError('First dimension of times must be equal to the number of channels')

        if voltages.ndim == 2 and voltages.shape[0] != len(channels):
            raise ValueError('First dimension of voltages must be equal to the number of channels')

        if voltages.shape[-1] != times.shape[-1]:
            ValueError('Different number of entries for times and voltages')

        return TablePulseTemplate(dict((channel, list(zip(times if times.ndim == 1 else times[i, :],
                                                          voltages if voltages.ndim == 1 else voltages[i, :])))
                                       for i, channel in enumerate(channels)))

    @staticmethod
    def from_entry_list(entry_list: List[Tuple],
                        channel_names: Optional[List[ChannelID]]=None, **kwargs) -> 'TablePulseTemplate':
        """Static constructor for a TablePulseTemplate where all channel's entries share the same times.

        :param entry_list: List of tuples of the form (t, v_1, ..., v_N[, interp])
        :param channel_names: Optional list of channel identifiers to use. Default is [0, ..., N-1]
        :param kwargs: Forwarded to TablePulseTemplate constructor
        :return: TablePulseTemplate with
        """
        # TODO: Better doc string
        def is_valid_interpolation_strategy(inter):
            return inter in TablePulseTemplate.interpolation_strategies or isinstance(inter, InterpolationStrategy)

        # determine number of channels
        max_len = max(len(data) for data in entry_list)
        min_len = min(len(data) for data in entry_list)

        if max_len - min_len > 1:
            raise ValueError('There are entries of contradicting lengths: {}'.format(set(len(t) for t in entry_list)))
        elif max_len - min_len == 1:
            num_chan = min_len - 1
        else:
            # figure out whether all last entries are interpolation strategies
            if all(is_valid_interpolation_strategy(interp) for *data, interp in entry_list):
                num_chan = min_len - 2
            else:
                num_chan = min_len - 1

        # insert default interpolation strategy key
        entry_list = [(t, *data, interp) if len(data) == num_chan else (t, *data, interp, 'default')
                      for t, *data, interp in entry_list]

        for *_, last_voltage, _ in entry_list:
            if last_voltage in TablePulseTemplate.interpolation_strategies:
                warnings.warn('{} is also an interpolation strategy name but handled as a voltage. Is it intended?'
                              .format(last_voltage), AmbiguousTablePulseEntry)

        if channel_names is None:
            channel_names = list(range(num_chan))
        elif len(channel_names) != num_chan:
            raise ValueError('Number of channel identifiers does not correspond to the number of channels.')

        parsed = {channel_name: [] for channel_name in channel_names}

        for time, *voltages, interp in entry_list:
            for channel_name, volt in zip(channel_names, voltages):
                parsed[channel_name].append((time, volt, interp))

        return TablePulseTemplate(parsed, **kwargs)

    @property
    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
        expressions = dict()
        for channel, channel_entries in self._entries.items():

            expr = 0
            for first_entry, second_entry in zip(channel_entries[:-1], channel_entries[1:]):
                substitutions = {'t0': ExpressionScalar(first_entry.t).sympified_expression, 'v0': ExpressionScalar(first_entry.v).sympified_expression,
                                 't1': ExpressionScalar(second_entry.t).sympified_expression, 'v1': ExpressionScalar(second_entry.v).sympified_expression}

                expr += first_entry.interp.integral.sympified_expression.subs(substitutions)
            expressions[channel] = ExpressionScalar(expr)

        return expressions


def concatenate(*table_pulse_templates: TablePulseTemplate, **kwargs) -> TablePulseTemplate:
    """Concatenate two or more table pulse templates"""
    first_template, *other_templates = table_pulse_templates

    entries = {channel: [] for channel in first_template.defined_channels}
    duration = ExpressionScalar(0)

    for i, template in enumerate(table_pulse_templates):
        if not isinstance(template, TablePulseTemplate):
            raise TypeError('Template number %d is not a TablePulseTemplate' % i)

        new_duration = duration + template.duration

        if template.defined_channels != first_template.defined_channels:
            raise ValueError('Template number %d has differing defined channels' % i,
                             first_template.defined_channels, template.defined_channels)

        for channel, channel_entries in template.entries.items():
            first_t, first_v, _ = channel_entries[0]
            if i > 0 and first_t != 0:
                if (first_v == 0) is False:
                    entries[channel].append((duration, first_v, 'hold'))

            for t, v, interp in channel_entries:
                entries[channel].append((duration.sympified_expression + t, v, interp))

            last_t, last_v, _ = channel_entries[-1]
            if i < len(other_templates) and last_t != new_duration:
                entries[channel].append((new_duration, last_v, TablePulseTemplate.interpolation_strategies['hold']))

        duration = new_duration

    return TablePulseTemplate(entries, **kwargs)


class ZeroDurationTablePulseTemplate(UserWarning):
    pass


class AmbiguousTablePulseEntry(UserWarning):
    pass



from typing import Dict, Any, Optional, Tuple
from collections import defaultdict

import numpy as np

from atsaverage.config import ScanlineConfiguration
from atsaverage.masks import CrossBufferMask, Mask

from qupulse.hardware.dacs.dac_base import DAC


class AlazarProgram:
    def __init__(self, masks=list(), operations=list(), total_length=None):
        self.masks = masks
        self.operations = operations
        self.total_length = total_length
    def __iter__(self):
        yield self.masks
        yield self.operations
        yield self.total_length


class AlazarCard(DAC):
    def __init__(self, card, config: Optional[ScanlineConfiguration]=None):
        self.__card = card

        self.__armed_program = None
        self.update_settings = True

        self.__definitions = dict()
        self.config = config

        self._mask_prototypes = dict()  # type: Dict

        self._registered_programs = defaultdict(AlazarProgram)  # type: Dict[str, AlazarProgram]

    @property
    def card(self) -> Any:
        return self.__card

    def _make_mask(self, mask_id: str, begins, lengths) -> Mask:
        if mask_id not in self._mask_prototypes:
            raise KeyError('Measurement window {} can not be converted as it is not registered.'.format(mask_id))

        hardware_channel, mask_type = self._mask_prototypes[mask_id]

        if np.any(begins[:-1]+lengths[:-1] > begins[1:]):
            raise ValueError('Found overlapping windows in begins')

        mask = CrossBufferMask()
        mask.identifier = mask_id
        mask.begin = begins
        mask.length = lengths
        mask.channel = hardware_channel
        return mask

    def register_measurement_windows(self,
                                     program_name: str,
                                     windows: Dict[str, Tuple[np.ndarray, np.ndarray]]) -> None:
        if not windows:
            self._registered_programs[program_name].masks = []
        total_length = 0
        for mask_id, (begins, lengths) in windows.items():

            sample_factor = self.config.captureClockConfiguration.numeric_sample_rate(self.__card.model) / 10**9

            begins = np.rint(begins*sample_factor).astype(dtype=np.uint64)
            lengths = np.floor(lengths*sample_factor).astype(dtype=np.uint64)

            sorting_indices = np.argsort(begins)
            begins = begins[sorting_indices]
            lengths = lengths[sorting_indices]

            windows[mask_id] = (begins, lengths)
            total_length = max(total_length, begins[-1]+lengths[-1])

        total_length = np.ceil(total_length/self.__card.minimum_record_size) * self.__card.minimum_record_size

        self._registered_programs[program_name].masks = [
            self._make_mask(mask_id, *window_begin_length)
            for mask_id, window_begin_length in windows.items()]
        self._registered_programs[program_name].total_length = total_length

    def register_operations(self, program_name: str, operations) -> None:
        self._registered_programs[program_name].operations = operations

    def arm_program(self, program_name: str) -> None:
        to_arm = self._registered_programs[program_name]
        if self.update_settings or self.__armed_program is not to_arm:
            config = self.config
            config.masks, config.operations, total_record_size = self._registered_programs[program_name]

            if len(config.operations) == 0:
                raise RuntimeError('No operations configured for program {}'.format(program_name))

            if not config.masks:
                if config.operations:
                    raise RuntimeError('Invalid configuration. Operations have no masks to work with')
                else:
                    return

            if config.totalRecordSize == 0:
                config.totalRecordSize = total_record_size
            elif config.totalRecordSize < total_record_size:
                raise ValueError('specified total record size is smaller than needed {} < {}'.format(config.totalRecordSize,
                                                                                                     total_record_size))
            
            old_aimed_buffer_size = config.aimedBufferSize
            
            # work around for measurments not working with one buffer
            if config.totalRecordSize < 5*config.aimedBufferSize:
                config.aimedBufferSize = config.totalRecordSize // 5

            config.apply(self.__card, True)

            # "Hide" work around from the user
            config.aimedBufferSize = old_aimed_buffer_size

            self.update_settings = False
            self.__armed_program = to_arm
        self.__card.startAcquisition(1)

    def delete_program(self, program_name: str) -> None:
        self._registered_programs.pop(program_name)
        # todo [2018-06-14]: what if program to delete is currently armed?

    def clear(self) -> None:
        self._registered_programs = dict()
        self.__armed_program = None

    @property
    def mask_prototypes(self) -> Dict[str, Tuple[int, str]]:
        return self._mask_prototypes

    def register_mask_for_channel(self, mask_id: str, hw_channel: int, mask_type='auto') -> None:
        """

        Args:
            mask_id: Identifier of the measurement windows
            hw_channel: Associated hardware channel (0, 1, 2, 3)
            mask_type: Either 'auto' or 'periodical
        """
        if hw_channel not in range(4):
            raise ValueError('{} is not a valid hw channel'.format(hw_channel))
        if mask_type not in ('auto', 'cross_buffer', None):
            raise NotImplementedError('Currently only can do cross buffer mask')
        self._mask_prototypes[mask_id] = (hw_channel, mask_type)

